# -*- coding: utf-8 -*-

"""
Python module to read in DR spectrometer data.  This module defines the following 
classes for storing the spectra found in a file:

Frame
  object that contains all data associated with a particular spectrometer frame.  
  The primary constituents of each frame are:
    * FrameHeader - the spectrometer frame header object and
    * FrameData   - the spectral data object.
Combined, these two objects contain all of the information found in the 
original spectrometer frame.

The functions defined in this module fall into two class:
  1. convert a frame in a file to a Frame object and
  2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object
For describing the format of data in the file, three function are provided:
  * getSampleRate - get the sample rate in the file
  * getFrameSize - get the total (header + data) frame size
  * getFFTsPerIntegration - get the number of FFT windows per integration
  * getTransformSize - get the FFT length
  * getIntegrationTime - get the integration time

.. note::
    This reader works with the most current version of the DR spectrometer data
    format as specified in the version 1.7.  To read data created with previous
    versions of the DR spectrometer, use LSL version 0.5.2.
    
.. versionchanged:: 1.0.1
    Added in new functions to help better describe the contents of a DR 
    spectrometer file.
"""

import copy
import numpy

from lsl.common import dp as dp_common
from lsl.reader.drx import filterCodes as drxFilterCodes
from lsl.reader._gofast import readDRSpec
from lsl.reader._gofast import syncError as gsyncError
from lsl.reader._gofast import eofError as geofError
from lsl.reader.errors import syncError, eofError

__version__ = '0.3'
__revision__ = '$Rev: 2465 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'readFrame', 'getDataProducts', 'containsLinearData',
           'containsStokesData', 'getSampleRate', 'getFrameSize', 'getFFTsPerIntegration', 
           'getTransformSize', 'getIntegrationTime', 'filterCodes', 
           '__version__', '__revision__', '__all__']

# List of filter codes and their corresponding sample rates in Hz.  
# .. note::
#		These are just the DRX filter codes
filterCodes = drxFilterCodes


class FrameHeader(object):
    """
    Class that stores the information found in the header of a DR spectrometer/DRX 
    frame.
    """
    
    def __init__(self, beam=0, format=0, decimation=None, timeOffset=None, nInts=None):
        self.beam = beam
        self.format = format
        self.decimation = decimation
        self.timeOffset = timeOffset
        
        if nInts is None:
            self.nInts = 0
        else:
            self.nInts = nInts
        
    def parseID(self):
        """
        Return the beam the frame corresponds to.
        """
        
        return self.beam
        
    def getDataProducts(self):
        """
        Return a list of data products contained in the file.
        
        .. versionadded:: 0.6.0
        """
        
        products = []
        
        # Linear
        if self.format & 0x01:
            products.append('XX')
        if self.format & 0x02:
            products.append('XY')
        if self.format & 0x04:
            products.append('YX')
        if self.format & 0x08:
            products.append('YY')
            
        # Stokes
        if self.format & 0x10:
            products.append('I')
        if self.format & 0x20:
            products.append('Q')
        if self.format & 0x40:
            products.append('U')
        if self.format & 0x80:
            products.append('V')
        
        return products
        
    def containsLinearData(self):
        """
        Return whether or not the frame contains linear polarization 
        products or not.
        
        .. versionadded:: 0.6.0
        """
        
        if self.format < 0x10:
            return True
        else:
            return False
            
    def containsStokesData(self):
        """
        Return whether or not the frame contains Stokes polarization
        parameters or not.
        
        .. versionadded:: 0.6.0
        """
        
        if self.format < 0x10:
            return False
        else:
            return True
        
    def getSampleRate(self):
        """
        Return the sample rate of the data in samples/second.
        """
        
        sampleRate = dp_common.fS / self.decimation
        return sampleRate
        
    def getFilterCode(self):
        """
        Function to convert the sample rate in Hz to a filter code.
        """
        
        sampleCodes = {}
        for key in filterCodes:
            value = filterCodes[key]
            sampleCodes[value] = key
            
        return sampleCodes[self.getSampleRate()]
        
    def getFFTsPerIntegration(self):
        """
        Return the number of FFT windows per integration.
        
        .. versionadded:: 1.0.1
        """
        
        return self.nInts


class FrameData(object):
    """
    Class that stores the information found in the data section of a DR spectrometer/
    DRX frame.
    
    .. versionchanged:: 0.5.3
        Added the saturations field to keep up with saturation counts.
        
    .. versionchanged:: 0.6.0
        The attributes that store the data are not defined until a frame is read in order
        to account for the fact that spectrometer files can store either linear or Stokes
        data.
    """
    
    def __init__(self, timeTag=None, tuningWords=None, fills=None, errors=None, saturations=None):
        self.timeTag = timeTag
        if tuningWords is None:
            self.tuningWords = [0, 0]
        else:
            self.tuningWords = tuningWords
        if fills is None:
            self.fills = [0, 0, 0, 0]
        else:
            self.fills = fills
        if errors is None:
            self.errors = [0, 0, 0, 0]
        else:
            self.errors = errors
        if saturations is None:
            self.saturations = [0, 0, 0, 0]
        else:
            self.saturations = saturations
            
    def getCentralFreq(self, which=None):
        """
        Function to set the central frequency of the DRX data in Hz.
        """
        
        if which is None:
            return [dp_common.fS * i / 2**32 for i in self.tuningWords]
        elif which == 1:
            return dp_common.fS * self.tuningWords[0] / 2**32
        elif which == 2:
            return dp_common.fS * self.tuningWords[1] / 2**32
        else:
            raise ValueError("Unknown tuning/polarization combination: '%i'" % which)
            
    def setGain(self, gain):
        """
        Function to set the gain of the DRX data.
        """

        self.gain = gain


class Frame(object):
    """
    Class that stores the information contained within a single DR spectrometer/
    DRX frame.  It's properties are FrameHeader and FrameDataLinear/FrameDataStokes
    objects.
    
    .. versionchanged:: 0.6.0
        By default the data contained with in a frame is normalized by the number of
        fills (header.fills parameter).  For data products that are a function of more
        than one primary input, i.e., XY* or I, the minimum fill of X and Y are used 
        for normalization.
    """
    
    def __init__(self, header=None, data=None):
        if header is None:
            self.header = FrameHeader()
        else:
            self.header = header
            
        if data is None:
            self.data = FrameData()
        else:
            self.data = data
            
        self.valid = True
        
    def parseID(self):
        """
        Convenience wrapper for the Frame.FrameHeader.parseID 
        function.
        """
        
        return self.header.parseID()
        
    def getDataProducts(self):
        """
        Convenience wrapper for the Frame.FrameHeder.getDataProducts
        function.
        """
        
        return self.header.getDataProducts()
        
    def containsLinearData(self):
        """
        Convenience wrapper for the Frame.FrameHeder.containsLinearData
        function.
        """
        
        return self.header.containsLinearData()
        
    def containsStokesData(self):
        """
        Convenience wrapper for the Frame.FrameHeder.containsStokesData
        function.
        """
        
        return self.header.containsStokesData()
        
    def getSampleRate(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getSampleRate 
        function.
        """
        
        return self.header.getSampleRate()
        
    def getFilterCode(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getFilterCode function.
        """
        
        return self.header.getFilterCode()
        
    def getFFTsPerIntegration(self):
        """
        Conveinence wrapper for the Frame.FrameHeader.getFFTsPerIntegration 
        function.
        
        .. versionadded:: 1.0.1
        """
        
        return self.header.getFFTsPerIntegration()
        
    def getTime(self):
        """
        Function to convert the time tag from samples since the UNIX epoch
        (UTC 1970-01-01 00:00:00) to seconds since the UNIX epoch.
        """
        
        seconds = (self.data.timeTag - self.header.timeOffset) / dp_common.fS
        
        return seconds
        
    def getCentralFreq(self, which=None):
        """
        Convenience wrapper for the Frame.FrameData.getCentralFreq function.
        """
        
        return self.data.getCentralFreq(which=which)
        
    def getTransformSize(self):
        """
        Find out what the transform size is.
        
        .. versionadded:: 1.0.1
        """
        
        p = self.getDataProducts()[0]
        return getattr(self.data, "%s0" % p, None).size
        
    def getIntegrationTime(self):
        """
        Return the integration time for data in seconds.
        
        .. versionadded:: 1.0.1
        """
        
        LFFT = self.getTransformSize()
        srate = self.getSampleRate()
        nInts = self.getFFTsPerIntegration()
        
        return nInts*LFFT/srate
        
    def setGain(self, gain):
        """
        Convenience wrapper for the Frame.FrameData.setGain function.
        """
        
        self.data.setGain(gain)
        
    def __add__(self, y):
        """
        Add the data sections of two frames together or add a number 
        to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame += y
        return newFrame
        
    def __iadd__(self, y):
        """
        In-place add the data sections of two frames together or add 
        a number to every element in the data section.
        """
        
        attrs = self.header.getDataProducts()
        
        for attrBase in attrs:
            for tuning in (0, 1):
                attr = "%s%i" % (attrBase, tuning)
                try:
                    temp = getattr(self.data, attr, None) + getattr(y.data, attr, None)
                except TypeError:
                    raise RuntimeError("Cannot add %s with %s" % (str(attrs), str(y.header.getDataProducts())))
                except AttributeError:
                    temp = getattr(self.data, attr, None) + numpy.float32(y)
                setattr(self.data, attr, temp)
            
        return self
        
    def __mul__(self, y):
        """
        Multiple the data sections of two frames together or multiply 
        a number to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame *= y
        return newFrame
            
    def __imul__(self, y):
        """
        In-place multiple the data sections of two frames together or 
        multiply a number to every element in the data section.
        """
        
        attrs = self.header.getDataProducts()
        
        for attrBase in attrs:
            for tuning in (0, 1):
                attr = "%s%i" % (attrBase, tuning)
                try:
                    temp = getattr(self.data, attr, None) * getattr(y.data, attr, None)
                except TypeError:
                    raise RuntimeError("Cannot multiply %s with %s" % (str(attrs), str(y.header.getDataProducts())))
                except AttributeError:
                    temp = getattr(self.data, attr, None) * numpy.float32(y)
                setattr(self.data, attr, temp)
            
        return self
            
    def __eq__(self, y):
        """
        Check if the time tags of two frames are equal or if the time
        tag is equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX == tY:
            return True
        else:
            return False
            
    def __ne__(self, y):
        """
        Check if the time tags of two frames are not equal or if the time
        tag is not equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX != tY:
            return True
        else:
            return False
            
    def __gt__(self, y):
        """
        Check if the time tag of the first frame is greater than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX > tY:
            return True
        else:
            return False
            
    def __ge__(self, y):
        """
        Check if the time tag of the first frame is greater than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX >= tY:
            return True
        else:
            return False
            
    def __lt__(self, y):
        """
        Check if the time tag of the first frame is less than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX < tY:
            return True
        else:
            return False
            
    def __le__(self, y):
        """
        Check if the time tag of the first frame is less than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX <= tY:
            return True
        else:
            return False
            
    def __cmp__(self, y):
        """
        Compare two frames based on the time tags.  This is helpful for 
        sorting things.
        """
        
        tX = self.data.timeTag
        tY = y.data.timeTag
        if tY > tX:
            return -1
        elif tX > tY:
            return 1
        else:
            return 0


def readFrame(filehandle, Gain=None, Verbose=False):
    """
    Function to read in a single DR spectrometer/DRX frame (header+data) and 
    store the contents as a Frame object.
    """
    
    # New Go Fast! (TM) method
    try:
        newFrame = readDRSpec(filehandle, Frame())
    except gsyncError:
        mark = filehandle.tell()
        raise syncError(location=mark)
    except geofError:
        raise eofError
        
    if Gain is not None:
        newFrame.setGain(Gain)
        
    return newFrame


def getDataProducts(filehandle):
    """
    Find out the data products contained in the file by looking at a frame.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Read in one frame
    newFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the data products
    return newFrame.header.getDataProducts()


def containsLinearData(filehandle):
    """
    Find out if the file contains linear polarization products or not.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Read in one frame
    newFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the verdict
    return newFrame.header.containsLinearData()


def containsStokesData(filehandle):
    """
    Find out if the file contains Stokes parameters or not.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Read in one frame
    newFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the verdict
    return newFrame.header.containsStokesData()


def getSampleRate(filehandle, nFrames=None, FilterCode=False):
    """
    Find out what the sampling rate/filter code is from a single observations.  
    By default, the rate in Hz is returned.  However, the corresponding filter 
    code can be returned instead by setting the FilterCode keyword to true.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Read in one frame
    newFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    if not FilterCode:
        return newFrame.getSampleRate()
    else:
        return newFrame.getFilterCode()


def getFrameSize(filehandle):
    """
    Find out what the frame size in a file is at the current file location.
    Returns the frame size in bytes.
    """
    
    cPos = filehandle.tell()
    frame = readFrame(filehandle)
    nPos = filehandle.tell()
    
    FrameSize = nPos - cPos
    filehandle.seek(cPos)
    
    return FrameSize


def getFFTsPerIntegration(filehandle):
    """
    Find out what the number of FFT windows per integration is at the 
    current file location.
    
    .. versionadded:: 1.0.1
    """
    
    cPos = filehandle.tell()
    frame = readFrame(filehandle)
    filehandle.seek(cPos)
    
    return frame.getFFTsPerIntegration()


def getTransformSize(filehandle):
    """
    Find out what the transform size in a file is at the current file 
    location.  
    """
    
    cPos = filehandle.tell()
    frame = readFrame(filehandle)
    filehandle.seek(cPos)
    
    return frame.getTransformSize()


def getIntegrationTime(filehandle):
    """
    Find out what the integration time is at the current file location.
    """
    
    cPos = filehandle.tell()
    frame = readFrame(filehandle)
    filehandle.seek(cPos)
    
    return frame.getIntegrationTime()



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Unit test for the lsl.common.idf module.
"""

import os
import pytz
import ephem
import tempfile
import unittest
from datetime import datetime, timedelta
try:
    import cStringIO as StringIO
except ImprotError:
    import StringIO

from lsl.common.paths import dataBuild as dataPath
from lsl.common import idf
from lsl.common.stations import lwa1, lwasv


__revision__ = "$Rev: 2774 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


drxFile = os.path.join(dataPath, 'tests', 'drx-idf.txt')
altFile = os.path.join(dataPath, 'tests', 'alt-idf.txt')
solFile = os.path.join(dataPath, 'tests', 'sol-idf.txt')
jovFile = os.path.join(dataPath, 'tests', 'jov-idf.txt')
sdfFile = os.path.join(dataPath, 'tests', 'drx-sdf.txt')


class idf_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.common.idf
    module."""
    
    testPath = None

    def setUp(self):
        """Create the temporary file directory."""

        self.testPath = tempfile.mkdtemp(prefix='test-idf-', suffix='.tmp')
        
    ### General ###
    
    def test_flat_projects(self):
        """Test single session/scans IDFs."""
        
        obs = idf.Observer('Test Observer', 99)
        targ = idf.DRX('Target', 'Target', '2019/1/1 00:00:00', '00:00:10', 0.0, 90.0, 40e6, 50e6, 6)
        sess = idf.Run('Test Session', 1, scans=targ)
        proj = idf.Project(obs, 'Test Project', 'COMTST', runs=sess)
        out = proj.render()
        
    def test_ucf_username(self):
        """Test setting the UCF username for auto-copy support."""
        
        obs = idf.Observer('Test Observer', 99)
        targ = idf.DRX('Target', 'Target', '2019/1/1 00:00:00', '00:00:10', 0.0, 90.0, 40e6, 50e6, 6)
        sess = idf.Run('Test Session', 1, scans=targ)
        sess.setDataReturnMethod('UCF')
        sess.setUCFUsername('test')
        proj = idf.Project(obs, 'Test Project', 'COMTST', runs=sess)
        out = proj.render()
        self.assertTrue(out.find('ucfuser:test') >= 0)
        
        obs = idf.Observer('Test Observer', 99)
        targ = idf.DRX('Target', 'Target', '2019/1/1 00:00:00', '00:00:10', 0.0, 90.0, 40e6, 50e6, 6)
        sess = idf.Run('Test Session', 1, scans=targ, comments='This is a comment')
        sess.setDataReturnMethod('UCF')
        sess.setUCFUsername('test/dir1')
        proj = idf.Project(obs, 'Test Project', 'COMTST', runs=sess)
        out = proj.render()
        self.assertTrue(out.find('ucfuser:test/dir1') >= 0)
        
    ### DRX - TRK_RADEC ###
    
    def test_drx_parse(self):
        """Test reading in a TRK_RADEC IDF file."""
        
        project = idf.parseIDF(drxFile)
        
        # Basic file structure
        self.assertEqual(len(project.runs), 1)
        self.assertEqual(len(project.runs[0].scans), 1)
        
        # Correlator setup
        self.assertEqual(project.runs[0].corr_channels, 256)
        self.assertAlmostEqual(project.runs[0].corr_inttime, 1.0, 6)
        self.assertEqual(project.runs[0].corr_basis, 'linear')
        
        # Observational setup - 1
        self.assertEqual(project.runs[0].scans[0].mode, 'TRK_RADEC')
        self.assertEqual(project.runs[0].scans[0].mjd,  58490)
        self.assertEqual(project.runs[0].scans[0].mpm,  74580000)
        self.assertEqual(project.runs[0].scans[0].dur,  7200000)
        self.assertEqual(project.runs[0].scans[0].freq1, 766958446)
        self.assertEqual(project.runs[0].scans[0].freq2, 1643482384)
        self.assertEqual(project.runs[0].scans[0].filter,   6)
        self.assertAlmostEqual(project.runs[0].scans[0].ra, 19.991210200, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].dec, 40.733916000, 6)
        
    def test_drx_update(self):
        """Test updating TRK_RADEC values."""
        
        project = idf.parseIDF(drxFile)
        project.runs[0].scans[0].setStart("MST 2011 Feb 23 17:00:15")
        project.runs[0].scans[0].setDuration(timedelta(seconds=15))
        project.runs[0].scans[0].setFrequency1(75e6)
        project.runs[0].scans[0].setFrequency2(76e6)
        project.runs[0].scans[0].setRA(ephem.hours('5:30:00'))
        project.runs[0].scans[0].setDec(ephem.degrees('+22:30:00'))
        
        self.assertEqual(project.runs[0].scans[0].mjd,  55616)
        self.assertEqual(project.runs[0].scans[0].mpm,  15000)
        self.assertEqual(project.runs[0].scans[0].dur,  15000)
        self.assertEqual(project.runs[0].scans[0].freq1, 1643482384)
        self.assertEqual(project.runs[0].scans[0].freq2, 1665395482)
        self.assertAlmostEqual(project.runs[0].scans[0].ra, 5.5, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].dec, 22.5, 6)
        
    def test_drx_write(self):
        """Test writing a TRK_RADEC IDF file."""
        
        project = idf.parseIDF(drxFile)
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        out = project.render()
        
    def test_drx_errors(self):
        """Test various TRK_RADEC IDF errors."""
        
        project = idf.parseIDF(drxFile)
        
        # Bad interferometer
        project.runs[0].stations = [lwa1,]
        self.assertFalse(project.validate())
        
        # Bad correlator channel count
        project.runs[0].corr_channels = 129
        self.assertFalse(project.validate())
        
        # Bad correlator integration time
        project.runs[0].corr_inttime = 1e-6
        self.assertFalse(project.validate())
        
        # Bad correlator output polarization basis
        project.runs[0].corr_basis = 'cats'
        self.assertFalse(project.validate())
        
        # Bad intent
        project.runs[0].scans[0].intent = 'cats'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad filter
        project.runs[0].scans[0].intent = 'Target'
        project.runs[0].scans[0].filter = 7
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.runs[0].scans[0].filter = 6
        project.runs[0].scans[0].frequency1 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        project.runs[0].scans[0].frequency1 = 38.0e6
        project.runs[0].scans[0].frequency2 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.runs[0].scans[0].frequency2 = 38.0e6
        project.runs[0].scans[0].duration = '96:00:00.000'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad pointing
        project.runs[0].scans[0].duration = '00:00:01.000'
        project.runs[0].scans[0].dec = -72.0
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - TRK_RADEC + Alternate phase centers###
    
    def test_drx_alt_parse(self):
        """Test reading in a TRK_RADEC IDF file with other phase centers."""
        
        project = idf.parseIDF(altFile)
        
        # Basic file structure
        self.assertEqual(len(project.runs), 1)
        self.assertEqual(len(project.runs[0].scans), 1)
        
        # Correlator setup
        self.assertEqual(project.runs[0].corr_channels, 256)
        self.assertAlmostEqual(project.runs[0].corr_inttime, 1.0, 6)
        self.assertEqual(project.runs[0].corr_basis, 'linear')
        
        # Observational setup - 1
        self.assertEqual(project.runs[0].scans[0].mode, 'TRK_RADEC')
        self.assertEqual(project.runs[0].scans[0].mjd,  58490)
        self.assertEqual(project.runs[0].scans[0].mpm,  74580000)
        self.assertEqual(project.runs[0].scans[0].dur,  7200000)
        self.assertEqual(project.runs[0].scans[0].freq1, 766958446)
        self.assertEqual(project.runs[0].scans[0].freq2, 1643482384)
        self.assertEqual(project.runs[0].scans[0].filter,   6)
        self.assertAlmostEqual(project.runs[0].scans[0].ra, 19.991210200, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].dec, 40.733916000, 6)
        
        # Phase center - 1
        self.assertEqual(len(project.runs[0].scans[0].alt_phase_centers), 2)
        self.assertAlmostEqual(project.runs[0].scans[0].alt_phase_centers[0].ra, 19.991310200, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].alt_phase_centers[0].dec, 40.733916000, 6)
        
        # Phase center - 2
        self.assertAlmostEqual(project.runs[0].scans[0].alt_phase_centers[1].ra, 19.991210200, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].alt_phase_centers[1].dec, 40.734016000, 6)
        
    def test_drx_alt_update(self):
        """Test updating TRK_RADEC values with other phase centers."""
        
        project = idf.parseIDF(altFile)
        project.runs[0].scans[0].setStart("MST 2011 Feb 23 17:00:15")
        project.runs[0].scans[0].setDuration(timedelta(seconds=15))
        project.runs[0].scans[0].setFrequency1(75e6)
        project.runs[0].scans[0].setFrequency2(76e6)
        project.runs[0].scans[0].setRA(ephem.hours('5:30:00'))
        project.runs[0].scans[0].setDec(ephem.degrees('+22:30:00'))
        project.runs[0].scans[0].alt_phase_centers[0].setRA(ephem.hours('5:35:00'))
        project.runs[0].scans[0].alt_phase_centers[1].setRA(ephem.hours('5:25:00'))
        
        self.assertEqual(project.runs[0].scans[0].mjd,  55616)
        self.assertEqual(project.runs[0].scans[0].mpm,  15000)
        self.assertEqual(project.runs[0].scans[0].dur,  15000)
        self.assertEqual(project.runs[0].scans[0].freq1, 1643482384)
        self.assertEqual(project.runs[0].scans[0].freq2, 1665395482)
        self.assertAlmostEqual(project.runs[0].scans[0].ra, 5.5, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].dec, 22.5, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].alt_phase_centers[0].ra, 5.583333, 6)
        self.assertAlmostEqual(project.runs[0].scans[0].alt_phase_centers[1].ra, 5.416667, 6)
        
    def test_drx_alt_write(self):
        """Test writing a TRK_RADEC IDF file with other phase centers."""
        
        project = idf.parseIDF(altFile)
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        out = project.render()
        
    def test_drx_alt_errors(self):
        """Test various TRK_RADEC IDF errors with other phase centers."""
        
        project = idf.parseIDF(altFile)
        
        # Bad interferometer
        project.runs[0].stations = [lwa1,]
        self.assertFalse(project.validate())
        
        # Bad correlator channel count
        project.runs[0].corr_channels = 129
        self.assertFalse(project.validate())
        
        # Bad correlator integration time
        project.runs[0].corr_inttime = 1e-6
        self.assertFalse(project.validate())
        
        # Bad correlator output polarization basis
        project.runs[0].corr_basis = 'cats'
        self.assertFalse(project.validate())
        
        # Bad intent
        project.runs[0].scans[0].intent = 'cats'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad filter
        project.runs[0].scans[0].intent = 'Target'
        project.runs[0].scans[0].filter = 7
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.runs[0].scans[0].filter = 6
        project.runs[0].scans[0].frequency1 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        project.runs[0].scans[0].frequency1 = 38.0e6
        project.runs[0].scans[0].frequency2 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.runs[0].scans[0].frequency2 = 38.0e6
        project.runs[0].scans[0].duration = '96:00:00.000'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad pointing
        project.runs[0].scans[0].duration = '00:00:01.000'
        project.runs[0].scans[0].dec = -72.0
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad alternate phase center
        project.runs[0].scans[0].dec = 40.733916000
        project.runs[0].scans[0].alt_phase_centers[0].dec = 45.0
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad alternate phase center intent
        project.runs[0].scans[0].alt_phase_centers[0].dec = 40.733916000
        project.runs[0].scans[0].alt_phase_centers[0].intent = 'cats'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Too many phase centers
        project.runs[0].scans[0].alt_phase_centers[0].intent = 'PhaseCal'
        for i in xrange(4):
            project.runs[0].scans[0].addAltPhaseCenter('test', 'Target', 19.991210200, 40.733916000)
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - TRK_SOL ###
    
    def test_sol_parse(self):
        """Test reading in a TRK_SOL IDF file."""
        
        project = idf.parseIDF(solFile)
        
        # Basic file structure
        self.assertEqual(len(project.runs), 1)
        self.assertEqual(len(project.runs[0].scans), 1)
        
        # Correlator setup
        self.assertEqual(project.runs[0].corr_channels, 512)
        self.assertAlmostEqual(project.runs[0].corr_inttime, 0.5, 6)
        self.assertEqual(project.runs[0].corr_basis, 'stokes')
        
        # Observational setup - 1
        self.assertEqual(project.runs[0].scans[0].mode, 'TRK_SOL')
        self.assertEqual(project.runs[0].scans[0].mjd,  58490)
        self.assertEqual(project.runs[0].scans[0].mpm,  74580000)
        self.assertEqual(project.runs[0].scans[0].dur,  7200000)
        self.assertEqual(project.runs[0].scans[0].freq1, 766958446)
        self.assertEqual(project.runs[0].scans[0].freq2, 1643482384)
        self.assertEqual(project.runs[0].scans[0].filter,   6)
        
    def test_sol_update(self):
        """Test updating TRK_SOL values."""
        
        project = idf.parseIDF(solFile)
        project.runs[0].scans[0].setStart("MST 2011 Feb 23 17:00:15")
        project.runs[0].scans[0].setDuration(timedelta(seconds=15))
        project.runs[0].scans[0].setFrequency1(75e6)
        project.runs[0].scans[0].setFrequency2(76e6)
        
        self.assertEqual(project.runs[0].scans[0].mjd,  55616)
        self.assertEqual(project.runs[0].scans[0].mpm,  15000)
        self.assertEqual(project.runs[0].scans[0].dur,  15000)
        self.assertEqual(project.runs[0].scans[0].freq1, 1643482384)
        self.assertEqual(project.runs[0].scans[0].freq2, 1665395482)
        
    def test_sol_write(self):
        """Test writing a TRK_SOL IDF file."""
        
        project = idf.parseIDF(solFile)
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        out = project.render()
        
    def test_sol_errors(self):
        """Test various TRK_SOL IDF errors."""
        
        project = idf.parseIDF(solFile)
        
        # Bad interferometer
        project.runs[0].stations = [lwa1,]
        self.assertFalse(project.validate())
        
        # Bad correlator channel count
        project.runs[0].corr_channels = 129
        self.assertFalse(project.validate())
        
        # Bad correlator integration time
        project.runs[0].corr_inttime = 1e-6
        self.assertFalse(project.validate())
        
        # Bad correlator output polarization basis
        project.runs[0].corr_basis = 'cats'
        self.assertFalse(project.validate())
        
        # Bad intent
        project.runs[0].scans[0].intent = 'cats'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad filter
        project.runs[0].scans[0].intent = 'Target'
        project.runs[0].scans[0].filter = 7
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.runs[0].scans[0].filter = 6
        project.runs[0].scans[0].frequency1 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        project.runs[0].scans[0].frequency1 = 38.0e6
        project.runs[0].scans[0].frequency2 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.runs[0].scans[0].frequency2 = 38.0e6
        project.runs[0].scans[0].duration = '96:00:00.000'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - TRK_JOV ###
    
    def test_jov_parse(self):
        """Test reading in a TRK_JOV IDF file."""
        
        project = idf.parseIDF(jovFile)
        
        # Basic file structure
        self.assertEqual(len(project.runs), 1)
        self.assertEqual(len(project.runs[0].scans), 1)
        
        # Correlator setup
        self.assertEqual(project.runs[0].corr_channels, 512)
        self.assertAlmostEqual(project.runs[0].corr_inttime, 0.1, 6)
        self.assertEqual(project.runs[0].corr_basis, 'circular')
        
        # Observational setup - 1
        self.assertEqual(project.runs[0].scans[0].mode, 'TRK_JOV')
        self.assertEqual(project.runs[0].scans[0].mjd,  58490)
        self.assertEqual(project.runs[0].scans[0].mpm,  74580000)
        self.assertEqual(project.runs[0].scans[0].dur,  3600000)
        self.assertEqual(project.runs[0].scans[0].freq1, 766958446)
        self.assertEqual(project.runs[0].scans[0].freq2, 1643482384)
        self.assertEqual(project.runs[0].scans[0].filter,   6)
        
    def test_jov_update(self):
        """Test updating TRK_JOV values."""
        
        project = idf.parseIDF(jovFile)
        project.runs[0].scans[0].setStart("MST 2011 Feb 23 17:00:15")
        project.runs[0].scans[0].setDuration(timedelta(seconds=15))
        project.runs[0].scans[0].setFrequency1(75e6)
        project.runs[0].scans[0].setFrequency2(76e6)
        
        self.assertEqual(project.runs[0].scans[0].mjd,  55616)
        self.assertEqual(project.runs[0].scans[0].mpm,  15000)
        self.assertEqual(project.runs[0].scans[0].dur,  15000)
        self.assertEqual(project.runs[0].scans[0].freq1, 1643482384)
        self.assertEqual(project.runs[0].scans[0].freq2, 1665395482)
        
    def test_jov_write(self):
        """Test writing a TRK_JOV IDF file."""
        
        project = idf.parseIDF(jovFile)
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        out = project.render()
        
    def test_jov_errors(self):
        """Test various TRK_JOV IDF errors."""
        
        project = idf.parseIDF(jovFile)
        
        # Bad interferometer
        project.runs[0].stations = [lwa1,]
        self.assertFalse(project.validate())
        
        # Bad correlator channel count
        project.runs[0].corr_channels = 129
        self.assertFalse(project.validate())
        
        # Bad correlator integration time
        project.runs[0].corr_inttime = 1e-6
        self.assertFalse(project.validate())
        
        # Bad correlator output polarization basis
        project.runs[0].corr_basis = 'cats'
        self.assertFalse(project.validate())
        
        # Bad intent
        project.runs[0].scans[0].intent = 'cats'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad filter
        project.runs[0].scans[0].intent = 'Target'
        project.runs[0].scans[0].filter = 7
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.runs[0].scans[0].filter = 6
        project.runs[0].scans[0].frequency1 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        project.runs[0].scans[0].frequency1 = 38.0e6
        project.runs[0].scans[0].frequency2 = 90.0e6
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.runs[0].scans[0].frequency2 = 38.0e6
        project.runs[0].scans[0].duration = '96:00:00.000'
        project.runs[0].scans[0].update()
        self.assertFalse(project.validate())
        
    ### Misc. ###
    
    def test_generate_sdfs(self):
        """Test generated SDFs from the IDF."""
        
        project = idf.parseIDF(drxFile)
        sdfs = project.generateSDFs()
        
    def test_auto_update(self):
        """Test project auto-update on render."""
        
        # Part 1 - frequency and duration
        project = idf.parseIDF(drxFile)
        project.runs[0].scans[0].frequency1 = 75e6
        project.runs[0].scans[0].duration = '00:01:31.000'
        
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        fh = open(os.path.join(self.testPath, 'idf.txt'), 'w')
        fh.write(project.render())
        fh.close()
        
        project = idf.parseIDF(os.path.join(self.testPath, 'idf.txt'))
        self.assertEqual(project.runs[0].scans[0].freq1, 1643482384)
        self.assertEqual(project.runs[0].scans[0].dur, 91000)
        
        # Part 2 - frequency and duration (timedelta)
        project = idf.parseIDF(drxFile)
        project.runs[0].scans[0].frequency1 = 75e6
        project.runs[0].scans[0].duration = timedelta(minutes=1, seconds=31, microseconds=1000)
        
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        fh = open(os.path.join(self.testPath, 'idf.txt'), 'w')
        fh.write(project.render())
        fh.close()
        
        project = idf.parseIDF(os.path.join(self.testPath, 'idf.txt'))
        self.assertEqual(project.runs[0].scans[0].freq1, 1643482384)
        self.assertEqual(project.runs[0].scans[0].dur, 91001)
        
        # Part 3 - frequency and start time
        project = idf.parseIDF(drxFile)
        project.runs[0].scans[0].frequency2 = 75e6
        project.runs[0].scans[0].start = "MST 2011 Feb 23 17:00:15"
        
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        fh = open(os.path.join(self.testPath, 'idf.txt'), 'w')		
        fh.write(project.render())
        fh.close()
        
        project = idf.parseIDF(os.path.join(self.testPath, 'idf.txt'))
        self.assertEqual(project.runs[0].scans[0].freq2, 1643482384)
        self.assertEqual(project.runs[0].scans[0].mjd,  55616)
        self.assertEqual(project.runs[0].scans[0].mpm,  15000)
        
        # Part 4 - frequency and start time (timedelta)
        project = idf.parseIDF(drxFile)
        _MST = pytz.timezone('US/Mountain')
        project.runs[0].scans[0].frequency2 = 75e6
        project.runs[0].scans[0].start = _MST.localize(datetime(2011, 2, 23, 17, 00, 30, 1000))
        
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        fh = open(os.path.join(self.testPath, 'idf.txt'), 'w')		
        fh.write(project.render())
        fh.close()
        
        project = idf.parseIDF(os.path.join(self.testPath, 'idf.txt'))
        self.assertEqual(project.runs[0].scans[0].freq2, 1643482384)
        self.assertEqual(project.runs[0].scans[0].mjd,  55616)
        self.assertEqual(project.runs[0].scans[0].mpm,  30001)
        
    def test_set_stations(self):
        """Test the set stations functionlity."""
        
        project = idf.parseIDF(drxFile)
        project.runs[0].setStations([lwasv, lwa1])
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        self.assertTrue(project.validate())
        
    def test_is_valid(self):
        """Test whether or not isValid works."""
        
        self.assertTrue(idf.isValid(drxFile))
        self.assertTrue(idf.isValid(solFile))
        self.assertTrue(idf.isValid(jovFile))
        
    def test_is_not_valid(self):
        """Test whether or not isValid works on LWA1 and IDF files."""
        
        self.assertFalse(idf.isValid(sdfFile))
        
    def test_username(self):
        """Test setting auto-copy parameters."""
        
        project = idf.parseIDF(drxFile)
        project.runs[0].setDataReturnMethod('UCF')
        project.runs[0].setUCFUsername('jdowell')
        # Fix for LWA-SV only going up to filter code 6
        for obs in project.runs[0].scans:
            obs.filter = 6
        out = project.render()
        
        self.assertTrue(out.find('Requested data return method is UCF') > 0)
        self.assertTrue(out.find('ucfuser:jdowell') > 0)
        
        fh = open(os.path.join(self.testPath, 'idf.txt'), 'w')		
        fh.write(out)
        fh.close()
        
        project = idf.parseIDF(os.path.join(self.testPath, 'idf.txt'))
        out = project.render()
        
        self.assertTrue(out.find('Requested data return method is UCF') > 0)
        self.assertTrue(out.find('ucfuser:jdowell') > 0)
        
    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class idf_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.idf units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(idf_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
This module implements a uniform DFT filter bank for use in calculating 
spectra as an alternative to a simple FFT.  The implementation here is based 
on:  http://www.scribd.com/doc/20561850/6/Polyphase-Filter-Coef%EF%AC%81cients

.. versionchanged:: 1.0.1
    Added support for using PyFFTW instead of NumPy for the FFTs
"""

import numpy
from fx import noWindow

try:
    import os
    import pickle
    import pyfftw
    
    from lsl.common.paths import data as dataPath
    
    # Enable the PyFFTW cache
    if not pyfftw.interfaces.cache.is_enabled():
        pyfftw.interfaces.cache.enable()
        pyfftw.interfaces.cache.set_keepalive_time(60)
        
    # Read in the wisdom (if it exists)
    wisdomFilename = os.path.join(dataPath, 'pyfftw-wisdom.pkl')
    if os.path.exists(wisdomFilename):
        fh = open(wisdomFilename, 'r')
        wisdom = pickle.load(fh)
        fh.close()
        
        pyfftw.import_wisdom(wisdom)
        useWisdom = True
    else:
        useWisdom = False
        
    usePyFFTW = True
    
except ImportError:
    usePyFFTW = False
    useWisdom = False


__version__ = '0.2'
__revision__ = '$Rev: 2390 $'
__all__ = ['fft', 'fft2', 'fft4', 'fft8', 'fft16', 'fft32', '__version__', '__revision__', '__all__']

def __filterCoeff(N, P):
    """
    Private function to generate the filter bank coefficients for N 
    channels using P taps.
    """

    t = numpy.arange(N*P)
    return numpy.sinc((t - N*P/2.0 + 0.5)/N)


def fft(signal, N, P=1, window=noWindow):
    """
    FFT-based poly-phase filter bank for creating N channels with P
    taps.  Optionally, a window function can be specified using the 
    'window' keyword.  See :mod:`lsl.correlator.fx.calcSpectra` for 
    details on using window functions.
    """
    
    filteredSignal = signal[0:N*P]*window(N*P)*__filterCoeff(N, P)
    
    if usePyFFTW and filteredSignal.dtype in (numpy.complex64, numpy.complex128):
        dd = filteredSignal.dtype
        di = pyfftw.empty_aligned(N, dtype=dd)
        do = pyfftw.empty_aligned(N, dtype=dd)
        
        forwardPlan = pyfftw.FFTW(di, do, direction='FFTW_FORWARD', flags=('FFTW_ESTIMATE',))
        
        for i in range(0, P):
            di[:] = filteredSignal[i*N:(i+1)*N]
            forwardPlan(di, do)
            try:
                fbOutput += do
            except NameError:
                fbOutput = do*1.0
                
    else:
        for i in range(0, P):
            fbTemp = numpy.fft.fft(filteredSignal[i*N:(i+1)*N])
            try:
                fbOutput += fbTemp
            except NameError:
                fbOutput = fbTemp*1.0
                
    return fbOutput

def fft2(signal, N, window=noWindow):
    """
    Sub-type of :mod:`lsl.correlator.filterbank.fft` that uses two taps.
    """

    return fft(signal, N, P=2, window=window)

def fft4(signal, N, window=noWindow):
    """
    Sub-type of :mod:`lsl.correlator.filterbank.fft` that uses four taps.
    """

    return fft(signal, N, P=4, window=window)

def fft8(signal, N, window=noWindow):
    """
    Sub-type of :mod:`lsl.correlator.filterbank.fft` that uses eight taps.
    """

    return fft(signal, N, P=8, window=window)

def fft16(signal, N, window=noWindow):
    """
    Sub-type of :mod:`lsl.correlator.filterbank.fft` that uses 16 taps.
    """

    return fft(signal, N, P=16, window=window)

def fft32(signal, N, window=noWindow):
    """
    Sub-type of :mod:`lsl.correlator.filterbank.fft` that uses 32 taps.
    """

    return fft(signal, N, P=32, window=window)




# -*- coding: utf-8 -*

# Python3 compatibility
from __future__ import print_function

import os
import imp
import sys
import glob
import tempfile
import unittest
import platform
import warnings
import subprocess
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from setuptools import setup, Command, Extension, Distribution, find_packages, __version__ as stv
from distutils import log
from distutils.command.build import build
try:
    # Attempt to use Cython for building extensions, if available
    from Cython.Distutils.build_ext import build_ext
    # Additionally, assert that the compiler module will load
    # also. Ref #1229.
    __import__('Cython.Compiler.Main')
except ImportError:
    from distutils.command.build_ext import build_ext
    
try:
    import numpy
except ImportError:
    pass
try:
    import multiprocessing
except ImportError:
    pass
import logging


def get_version():
    """Read the VERSION file and return the version number as a string."""

    return open('VERSION').read().strip()


def get_description(filename):
    """Read in a README-type file and return the contents of the DESCRIPTION
    section."""

    desc = ''
    fh = open(filename, 'r')
    lines = fh.readlines()
    fh.close()

    inDescription = False
    for line in lines:
        if line.find('DESCRIPTION') == 0:
            inDescription = True
            continue
        if line.find('REQUIREMENTS') == 0:
            inDescription = False
            break
        if inDescription:
            if line[:3] == '---':
                continue
            desc = ''.join([desc, line])

    return desc


def get_openmp():
    """Try to compile/link an example program to check for OpenMP support.
    
    Based on:
    1) http://stackoverflow.com/questions/16549893/programatically-testing-for-openmp-support-from-a-python-setup-script
    2) https://github.com/lpsinger/healpy/blob/6c3aae58b5f3281e260ef7adce17b1ffc68016f0/setup.py
    """
    
    import shutil
    from distutils import sysconfig
    from distutils import ccompiler
    compiler = ccompiler.new_compiler()
    sysconfig.get_config_vars()
    sysconfig.customize_compiler(compiler)
    cc = compiler.compiler
    
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    
    fh = open('test.c', 'w')
    fh.write(r"""#include <omp.h>
#include <stdio.h>
int main(void) {
#pragma omp parallel
printf("Hello from thread %d, nthreads %d\n", omp_get_thread_num(), omp_get_num_threads());
return 0;
}
""")
    fh.close()
    
    ccmd = []
    ccmd.extend( cc )
    ccmd.extend( ['-fopenmp', 'test.c', '-o test', '-lgomp'] )
    try:
        output = subprocess.check_call(ccmd)
        outCFLAGS = ['-fopenmp',]
        outLIBS = ['-lgomp',]
        
    except subprocess.CalledProcessError:
        print("WARNING:  OpenMP does not appear to be supported by %s, disabling" % cc[0])
        outCFLAGS = []
        outLIBS = []
        
    finally:
        os.chdir(curdir)
        shutil.rmtree(tmpdir)
        
    return outCFLAGS, outLIBS


def get_fftw():
    """Use pkg-config (if installed) to figure out the C flags and linker flags
    needed to compile a C program with FFTW3 (floating point version).  If FFTW3 
    cannot be found via pkg-config, some 'sane' values are returned."""
    
    try:
        subprocess.check_call(['pkg-config', 'fftw3f', '--exists'])
        
        p = subprocess.Popen(['pkg-config', 'fftw3f', '--modversion'], stdout=subprocess.PIPE)
        outVersion = p.communicate()[0].rstrip().split()
        
        p = subprocess.Popen(['pkg-config', 'fftw3f', '--cflags'], stdout=subprocess.PIPE)
        outCFLAGS = p.communicate()[0].rstrip().split()
        try:
            outCFLAGS = [str(v, 'utf-8') for v in outCFLAGS]
        except TypeError:
            pass
        
        p = subprocess.Popen(['pkg-config', 'fftw3f', '--libs'], stdout=subprocess.PIPE)
        outLIBS = p.communicate()[0].rstrip().split()
        try:
            outLIBS = [str(v, 'utf-8') for v in outLIBS]
        except TypeError:
            pass
            
        if len(outVersion) > 0:
            print("Found FFTW3, version %s" % outVersion[0])
            
    except (OSError, subprocess.CalledProcessError):
        print("WARNING:  FFTW3 cannot be found, using defaults")
        outCFLAGS = []
        outLIBS = ['-lfftw3f', '-lm']
        
    return outCFLAGS, outLIBS


def get_atlas():
    """Use NumPy's system_info module to find the location of ATLAS (and 
    cblas)."""
    
    from numpy.distutils.system_info import get_info
    
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",category=DeprecationWarning)
        sys.stdout = StringIO()
        atlas_info = get_info('atlas_blas', notfound_action=2)
        sys.stdout = sys.__stdout__
        
    atlas_version = ([v[3:-3] for k,v in atlas_info.get('define_macros',[])
                    if k == 'ATLAS_INFO']+[None])[0]
    if atlas_version:
        print("Found ATLAS, version %s" % atlas_version)
        
    outCFLAGS = ['-I%s' % idir for idir in atlas_info['include_dirs']]
    outLIBS = ['-L%s' % ldir for ldir in atlas_info['library_dirs']]
    outLIBS.extend(['-l%s' % lib for lib in atlas_info['libraries']])
    
    return outCFLAGS, outLIBS


def write_version_info():
    """Write the version info to a module in LSL."""
    
    lslVersion = get_version()
    shortVersion = '.'.join(lslVersion.split('.')[:2])
    
    contents = """# -*- coding: utf-8 -*-
# This file is automatically generated by setup.py

import os
import glob
import hashlib

def _get_md5(filename, blockSize=262144):
    with open(filename, 'rb') as fh:
        m = hashlib.md5()
        while True:
            block = fh.read(blockSize)
            if len(block) == 0:
                break
            m.update(block)
            
    return m.hexdigest()

def get_fingerprint():
    \"\"\"
    Return a 'fingerprint' of the current LSL module state that is useful for
    tracking if the module has changed.
    \"\"\"
    
    filenames = []
    for ext in ('.py', '.so'):
        filenames.extend( glob.glob(os.path.join(os.path.dirname(__file__), '*%%s' %% ext)) )
        filenames.extend( glob.glob(os.path.join(os.path.dirname(__file__), '*', '*%%s' %% ext)) )
        
    m = hashlib.md5()
    for filename in filenames:
        m.update(_get_md5(filename))
    return m.hexdigest()

version = '%s'
full_version = '%s'
short_version = '%s'

""" % (lslVersion, lslVersion, shortVersion)
    
    fh = open('lsl/version.py', 'w')
    fh.write(contents)
    fh.close()
    
    return True


# Get the FFTW flags/libs and manipulate the flags and libraries for 
# correlator._core appropriately.  This will, hopefully, fix the build
# problems on Mac
class lsl_build(build):
    user_options = build.user_options \
                   + [('with-atlas=', None, 'Installation path for ATLAS'),] \
                   + [('with-fftw=', None, 'Installation path for FFTW'),]
    
    def initialize_options(self, *args, **kwargs):
        build.initialize_options(self, *args, **kwargs)
        self.with_atlas = None
        self.with_fftw = None
        
    def finalize_options(self, *args, **kwargs):
        build.finalize_options(self, *args, **kwargs)
        
        if self.distribution.has_ext_modules():
            ## Grab the 'build_ext' command
            beco = self.distribution.get_command_obj('build_ext')
            
            ## Grab the ATLAS flags
            if self.with_atlas is not None:
                beco.with_atlas = self.with_atlas
                
            ## Grab the FFTW flags
            if self.with_fftw is not None:
                beco.with_fftw = self.with_fftw


class lsl_build_ext(build_ext):
    user_options = build_ext.user_options \
                   + [('with-atlas=', None, 'Installation path for ATLAS'),] \
                   + [('with-fftw=', None, 'Installation path for FFTW'),]
    
    def initialize_options(self, *args, **kwargs):
        build_ext.initialize_options(self, *args, **kwargs)
        self.with_atlas = None
        self.with_fftw = None
        
    def finalize_options(self, *args, **kwargs):
        build_ext.finalize_options(self, *args, **kwargs)
        self.verbose = True
        
        ## Grab the OpenMP flags
        openmpFlags, openmpLibs = get_openmp()
        
        ## Grab the ATLAS flags
        if self.with_atlas is not None:
            atlasFlags = ['-I%s/include' % self.with_atlas,]
            atlasLibs = ['-L%s/lib' % self.with_atlas, '-lf77blas', '-lcblas', '-latlas']
        else:
            atlasFlags, atlasLibs = get_atlas()
            
        ## Grab the FFTW flags
        if self.with_fftw is not None:
            fftwFlags = ['-I%s/include' % self.with_fftw,]
            fftwLibs = ['-L%s/lib' % self.with_fftw, '-lfftw3f']
        else:
            fftwFlags, fftwLibs = get_fftw()
            
        ## Update the extensions with the additional compilier/linker flags
        for ext in self.extensions:
            ### Compiler flags
            for cflags in (openmpFlags, atlasFlags, fftwFlags):
                try:
                    ext.extra_compile_args.extend( cflags )
                except TypeError:
                    ext.extra_compile_args = cflags
            ### Linker flags
            for ldflags in (openmpLibs, atlasLibs, fftwLibs):
                try:
                    ext.extra_link_args.extend( ldflags )
                except TypeError:
                    ext.extra_link_args = ldflags
                    
        ## HACK: Update the log verbosity - for some reason this gets set to 
        ##       WARN when I replace build_ext
        log.set_threshold(min([log.INFO, log._global_log.threshold]))


coreExtraFlags = ['-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION']
coreExtraLibs = []


# Create the list of extension modules.  We do this here so that we can turn 
# off the DRSU direct module for non-linux system
ExtensionModules = [Extension('reader._gofast', ['lsl/reader/gofast.c', 'lsl/reader/tbw.c', 'lsl/reader/tbn.c', 'lsl/reader/drx.c', 'lsl/reader/drspec.c', 'lsl/reader/vdif.c', 'lsl/reader/tbf.c', 'lsl/reader/cor.c'], include_dirs=[numpy.get_include()], extra_compile_args=['-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION', '-funroll-loops']),
            Extension('common._fir', ['lsl/common/fir.c'], include_dirs=[numpy.get_include()], libraries=['m'], extra_compile_args=coreExtraFlags, extra_link_args=coreExtraLibs),
            Extension('correlator._spec', ['lsl/correlator/spec.c'], include_dirs=[numpy.get_include()], libraries=['m'], extra_compile_args=coreExtraFlags, extra_link_args=coreExtraLibs), 
            Extension('correlator._stokes', ['lsl/correlator/stokes.c'], include_dirs=[numpy.get_include()], libraries=['m'], extra_compile_args=coreExtraFlags, extra_link_args=coreExtraLibs),
            Extension('correlator._core', ['lsl/correlator/core.c'], include_dirs=[numpy.get_include()], libraries=['m'], extra_compile_args=coreExtraFlags, extra_link_args=coreExtraLibs), 
            Extension('sim._simfast', ['lsl/sim/simfast.c', 'lsl/sim/const.c', 'lsl/sim/j1.c', 'lsl/sim/polevl.c', 'lsl/sim/mtherr.c', 'lsl/sim/sf_error.c'], include_dirs=[numpy.get_include()], libraries=['m'], extra_compile_args=coreExtraFlags, extra_link_args=coreExtraLibs), 
            Extension('misc._wisdom', ['lsl/misc/wisdom.c'],include_dirs=[numpy.get_include()], libraries=['m'], extra_compile_args=coreExtraFlags, extra_link_args=coreExtraLibs), ]

# Update the version information
write_version_info()

# Setup the INSTALL_REQUIRES list and the EXTRA_REQUIRE dictionary and then
# deal with all of the things that can go wrong with old versions of 
# setuptools
#  From:  https://hynek.me/articles/conditional-python-dependencies/
INSTALL_REQUIRES = ['pyfits>=3.2', 'numpy>=1.2', 'scipy>=0.7', 'pyephem>=3.7.5.3', 'aipy>=1.0', 'pytz>=2011k']
EXTRA_REQUIRE = {}
if int(stv.split('.')[0], 10) < 18:
    if sys.version_info[:2] < (2,7):
        INSTALL_REQUIRES.append("argparse")
        INSTALL_REQUIRES.append("unittest2")
else:
    EXTRA_REQUIRE[":python_version<'2.7'"] = ["argparse", "unittest2"]

setup(
    cmdclass = {'build': lsl_build, 'build_ext': lsl_build_ext}, 
    name = "lsl", 
    version = get_version(), 
    description = "LWA Software Library", 
    author = "Jayce Dowell", 
    author_email = "jdowell@unm.edu", 
    url = "https://fornax.phys.unm.edu/lwa/trac/", 
    long_description = get_description('README.md'), 
    license = 'GPL',
    classifiers = ['Development Status :: 5 - Production/Stable',
                  'Intended Audience :: Developers',
                  'Intended Audience :: Science/Research',
                  'License :: OSI Approved :: GNU General Public License (GPL)',
                  'Topic :: Scientific/Engineering :: Astronomy',
                  'Programming Language :: Python :: 2',
                  'Programming Language :: Python :: 2.6',
                  'Programming Language :: Python :: 2.7',
                  'Operating System :: MacOS :: MacOS X',
                  'Operating System :: POSIX :: Linux'],
    packages = find_packages(), 
    scripts = glob.glob('scripts/*.py'), 
    python_requires='>=2.6, <3',
    setup_requires = ['numpy>=1.2'], 
    install_requires = INSTALL_REQUIRES, 
    extras_require = EXTRA_REQUIRE, 
    include_package_data = True,  
    ext_package = 'lsl', 
    ext_modules = ExtensionModules,
    zip_safe = False,  
    test_suite = "tests"
) 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import pytz
import getopt
import argparse

from datetime import datetime, timedelta

from lsl.common import stations
from lsl.astro import utcjd_to_unix, MJD_OFFSET
from lsl.common import metabundle, metabundleADP


__version__ = "0.2"
__revision__ = "$Rev: 941 $"

# Date/time manipulation
_UTC = pytz.utc
formatString = '%Y/%m/%d %H:%M:%S.%f %Z'


def getObsStartStop(obs):
    """
    Given an observation, get the start and stop times (returned as a two-
    element tuple).
    """
    
    # UNIX timestamp for the start
    tStart = utcjd_to_unix(obs.mjd + MJD_OFFSET)
    tStart += obs.mpm / 1000.0
    
    # UNIX timestamp for the stop
    tStop = tStart +  obs.dur / 1000.0
    
    # Conversion to a timezone-aware datetime instance
    tStart = _UTC.localize( datetime.utcfromtimestamp(tStart) )
    tStop  = _UTC.localize( datetime.utcfromtimestamp(tStop ) )
    
    # Return
    return tStart, tStop


def main(args):
    # Get the site and observer
    site = stations.lwa1
    observer = site.getObserver()
    
    # Filenames in an easier format
    inputTGZ  = args.filename
    
    # Parse the input file and get the dates of the observations.  Be default 
    # this is for LWA1 but we switch over to LWA-SV if an error occurs.
    try:
        # LWA1
        project = metabundle.getSessionDefinition(inputTGZ)
        obsImpl = metabundle.getObservationSpec(inputTGZ)
        fileInfo = metabundle.getSessionMetaData(inputTGZ)
        aspConfigB = metabundle.getASPConfigurationSummary(inputTGZ, which='Beginning')
        aspConfigE = metabundle.getASPConfigurationSummary(inputTGZ, which='End')
    except:
        # LWA-SV
        ## Site changes
        site = stations.lwasv
        observer = site.getObserver()
        ## Try again
        project = metabundleADP.getSessionDefinition(inputTGZ)
        obsImpl = metabundleADP.getObservationSpec(inputTGZ)
        fileInfo = metabundleADP.getSessionMetaData(inputTGZ)
        aspConfigB = metabundleADP.getASPConfigurationSummary(inputTGZ, which='Beginning')
        aspConfigE = metabundleADP.getASPConfigurationSummary(inputTGZ, which='End')
        
    nObs = len(project.sessions[0].observations)
    tStart = [None,]*nObs
    for i in xrange(nObs):
        tStart[i]  = utcjd_to_unix(project.sessions[0].observations[i].mjd + MJD_OFFSET)
        tStart[i] += project.sessions[0].observations[i].mpm / 1000.0
        tStart[i]  = datetime.utcfromtimestamp(tStart[i])
        tStart[i]  = _UTC.localize(tStart[i])
    
    # Get the LST at the start
    observer.date = (min(tStart)).strftime('%Y/%m/%d %H:%M:%S')
    lst = observer.sidereal_time()
    
    # Report on the file
    print "Filename: %s" % inputTGZ
    print " Project ID: %s" % project.id
    print " Session ID: %i" % project.sessions[0].id
    print " Observations appear to start at %s" % (min(tStart)).strftime(formatString)
    print " -> LST at %s for this date/time is %s" % (site.name, lst)
    
    lastDur = project.sessions[0].observations[nObs-1].dur
    lastDur = timedelta(seconds=int(lastDur/1000), microseconds=(lastDur*1000) % 1000000)
    sessionDur = max(tStart) - min(tStart) + lastDur
    
    print " "
    print " Total Session Duration: %s" % sessionDur
    print " -> First observation starts at %s" % min(tStart).strftime(formatString)
    print " -> Last observation ends at %s" % (max(tStart) + lastDur).strftime(formatString)
    if project.sessions[0].observations[0].mode not in ('TBW', 'TBN'):
        drspec = 'No'
        if project.sessions[0].spcSetup[0] != 0 and project.sessions[0].spcSetup[1] != 0:
            drspec = 'Yes'
        drxBeam = project.sessions[0].drxBeam
        if drxBeam < 1:
            drxBeam = "MCS decides"
        else:
            drxBeam = "%i" % drxBeam
        print " DRX Beam: %s" % drxBeam
        print " DR Spectrometer used? %s" % drspec
        if drspec == 'Yes':
            print " -> %i channels, %i windows/integration" % tuple(project.sessions[0].spcSetup)
    else:
        tbnCount = 0
        tbwCount = 0
        for obs in project.sessions[0].observations:
            if obs.mode == 'TBW':
                tbwCount += 1
            else:
                tbnCount += 1
        if tbwCount > 0 and tbnCount == 0:
            print " Transient Buffer Mode: TBW"
        elif tbwCount == 0 and tbnCount > 0:
            print " Transient Buffer Mode: TBN"
        else:
            print " Transient Buffer Mode: both TBW and TBN"
    print " "
    print "File Information:"
    for obsID in fileInfo.keys():
        print " Obs. #%i: %s" % (obsID, fileInfo[obsID]['tag'])
    
    print " "
    print "ASP Configuration:"
    print '  Beginning'
    for k in aspConfigB.keys():
        print '    %s: %i' % (k, aspConfigB[k])
    print '  End'
    for k in aspConfigE.keys():
        print '    %s: %i' % (k, aspConfigE[k])
        
    print " "
    print " Number of observations: %i" % nObs
    print " Observation Detail:"
    for i in xrange(nObs):
        currDur = project.sessions[0].observations[i].dur
        currDur = timedelta(seconds=int(currDur/1000), microseconds=(currDur*1000) % 1000000)
        
        print "  Observation #%i" % (i+1,)
        currObs = None
        for j in xrange(len(obsImpl)):
            if obsImpl[j]['obsID'] == i+1:
                currObs = obsImpl[j]
                break
                
        ## Basic setup
        print "   Target: %s" % project.sessions[0].observations[i].target
        print "   Mode: %s" % project.sessions[0].observations[i].mode
        print "   Start:"
        print "    MJD: %i" % project.sessions[0].observations[i].mjd
        print "    MPM: %i" % project.sessions[0].observations[i].mpm
        print "    -> %s" % getObsStartStop(project.sessions[0].observations[i])[0].strftime(formatString)
        print "   Duration: %s" % currDur
        
        ## DP setup
        if project.sessions[0].observations[i].mode not in ('TBW',):
            print "   Tuning 1: %.3f MHz" % (project.sessions[0].observations[i].frequency1/1e6,)
        if project.sessions[0].observations[i].mode not in ('TBW', 'TBN'):
            print "   Tuning 2: %.3f MHz" % (project.sessions[0].observations[i].frequency2/1e6,)
        if project.sessions[0].observations[i].mode not in ('TBW',):
            print "   Filter code: %i" % project.sessions[0].observations[i].filter
        if currObs is not None:
            if project.sessions[0].observations[i].mode not in ('TBW',):
                if project.sessions[0].observations[i].mode == 'TBN':
                    print "   Gain setting: %i" % currObs['tbnGain']
                else:
                    print "   Gain setting: %i" % currObs['drxGain']
        else:
            print "   WARNING: observation specification not found for this observation"
            
        ## Comments/notes
        print "   Observer Comments: %s" % project.sessions[0].observations[i].comments


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='display information about an LWA metadata tarball', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('filename', type=str, 
                        help='metadata file to display')
    args = parser.parse_args()
    main(args)
    


# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function

"""
Module for building and saving LSL-specific FFTW and PyFFTW wisdom.

.. versionadded:: 1.0.1
"""

import os
import numpy
import pickle
from datetime import datetime

from lsl.common.paths import data as dataPath
from lsl.common.busy import BusyIndicator
from lsl.misc import _wisdom


__version__ = "0.2"
__revision__ = "$Rev: 2454 $"
__all__ = ["make", "show", "__version__", "__revision__", "__all__"]


# Path to the LSL-specific FFTW wisdom file
_wisdomFilenameFFTW = os.path.join(dataPath, 'fftw_wisdom.txt')

# Path to the LSL-specific PyFFTW wisdom file
_wisdomFilenamePyFFTW = os.path.join(dataPath, 'pyfftw_wisdom.pkl')


def _makeFFTW():
    """
    Build a new set of LSL-specific FFTW wisdom.
    """
    
    bi = BusyIndicator(message="Building FFTW wisdom")
    bi.start()
    
    _wisdom.buildWisdom(_wisdomFilenameFFTW)
    
    bi.stop()
    
    return True


def _makePyFFTW():
    """
    Build a new set of LSL-specific PyFFTW wisdom.
    """
    
    MAXTRANSFORM = 262144
    
    try:
        import pyfftw
        
        bi = BusyIndicator(message="Building PyFFTW wisdom")
        bi.start()
        
        # Enable the PyFFTW cache
        if not pyfftw.interfaces.cache.is_enabled():
            pyfftw.interfaces.cache.enable()
            pyfftw.interfaces.cache.set_keepalive_time(60)
            
        fftFunction = lambda x: pyfftw.interfaces.numpy_fft.fft(x, planner_effort='FFTW_PATIENT')
        ifftFunction = lambda x: pyfftw.interfaces.numpy_fft.ifft(x, planner_effort='FFTW_PATIENT', overwrite_input=True)
        
        # Setup
        fftlen = 2
        while fftlen <= MAXTRANSFORM:
            data = numpy.ones(fftlen, dtype=numpy.complex64)
            fftFunction(data)
            ifftFunction(data)
            
            fftlen *= 2
            
        fftlen = 10
        while fftlen <= MAXTRANSFORM:
            data = numpy.ones(fftlen, dtype=numpy.complex64)
            fftFunction(data)
            ifftFunction(data)
            
            fftlen *= 10
            
        fh = open(_wisdomFilenamePyFFTW, 'wb')
        pickle.dump(pyfftw.export_wisdom(), fh)
        fh.close()
        
        bi.stop()
        
    except ImportError:
        print("PyFFTW is not installed, skipping")
        return False
        
    return True


def make(FFTW=True, PyFFTW=True):
    """
    Build a new set of LSL-specific FFTW and, optionally, PyFFTW wisdom.
    """
    
    # FFTW
    if FFTW:
        _makeFFTW()
        
    # PyFFTW
    if PyFFTW:
        _makePyFFTW()
        
    return True


def _showFFTW():
    """
    List information about the current LSL-specific FFTW wisdom.
    """
    
    if not os.path.exists(_wisdomFilenameFFTW):
        print("No LSL-specific FFTW wisdom file found, consider running make()")
        return False
        
    fh = open(_wisdomFilenameFFTW, 'r')
    lines = fh.readlines()
    fh.close()
    
    print("LSL FFTW Wisdom:")
    print(" Lines: %i" % len(lines))
    print(" Size: %i bytes" % os.path.getsize(_wisdomFilenameFFTW))
    print(" Last Modified: %s" % datetime.utcfromtimestamp(os.stat(_wisdomFilenameFFTW)[8]))
    
    return True


def _showPyFFTW():
    """
    List information about the current LSL-specific FFTW wisdom.
    """
    
    if not os.path.exists(_wisdomFilenamePyFFTW):
        print("No LSL-specific PyFFTW wisdom file found, consider running make()")
        return False
        
    try:
        import pyfftw
        
        fh = open(_wisdomFilenamePyFFTW, 'r')
        d,s,ld = pickle.load(fh)
        fh.close()
        
        d = d.split('\n')
        s = s.split('\n')
        ld = ld.split('\n')
        
        print("LSL PyFFTW Wisdom:")
        print(" Lines: %i (single) %i (double) %i (long double)" % (len(s), len(d), len(ld)))
        print(" Size: %i bytes" % os.path.getsize(_wisdomFilenamePyFFTW))
        print(" Last Modified: %s" % datetime.utcfromtimestamp(os.stat(_wisdomFilenamePyFFTW)[8]))
        
    except ImportError:
        print("PyFFTW is not installed, skipping")
        return False
    
    return True


def show(FFTW=True, PyFFTW=True):
    """
    List information about the current LSL-specific FFTW and PyFFTW wisdom.
    """
    
    # FFTW
    if FFTW:
        _showFFTW()
        
    # PyFFTW
    if PyFFTW:
        _showPyFFTW()
        
    return True



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Classes and methods to model sky brightness and visibility.

.. versionchanged:: 1.2.0
    Removed the orignal SkyMap class that uses LFmap at 73.9 MHz
    Updated SkyMapGSM to use the actual GSM rather than a fit to the GSM
    Added a new SkyMapLFSM that uses the 5.1 degree resolution LFSM
"""

### This module handles the skymaps from the 74 MHz skymap.
### David Munton, ARL:UT Jan 2007

import os
from numpy import pi, float32, log, exp, log10, sin, cos, arcsin, arccos, empty, arange, compress, clip, where, load

import pyfits
from scipy.interpolate import interp1d

from lsl import astro
from lsl.common.paths import data as dataPath

__version__   = '0.3'
__revision__ = '$Rev: 2457 $'
__author__    = 'J. York'
__maintainer__ = 'Jayce Dowell'


### This code is the base class for the sky map. It takes as input a skymap file name and frequency to
### which the skymap corresponds.  It has the following methods:
###     _init_ - takes the array coordinate filename as an input argument.
###     NormalizePower - Converts the skymap powers (in Kelvin radiated into 4 pi ster) into a power 
###                      seen at the antenna.
###     ComputeTotalPowerFromSky - Sums the power for all sources in the sky
###     ScaleSourcePowerstoFrequency - Scales the skymap from the base 73.8 MHz to the desired frequency.

class SkyMapError(Exception):
    """
    Base class for SkyMap exceptions.
    """
    
    def __init__(self, *args, **kwds):
        super(SkyMapError, self).__init__(*args, **kwds)


class SkyMapLoadError(SkyMapError):
    """
    Exception raised when a SkyMap fails to load.
    """
    
    def __init__(self, *args, **kwds):
        super(SkyMapLoadError, self).__init__(*args, **kwds)


class SkyMapPowerError(SkyMapError):
    """
    Exception raised when an operation on a normalized SkyMap fails to 
    compute (such as summing or multiplying over an empty data set).
    """
    
    def __init__(self, *args, **kwds):
        super(SkyMapPowerError, self).__init__(*args, **kwds)


class SkyMapGSM(object):
    """
    Extension of the SkyMap class to use the Global Sky Model.
    
    For more information on the Global Sky Model, see: http://space.mit.edu/~angelica/gsm/index.html
    
    .. note:: This class uses a slightly different interpolation method than
              the original GSM and introduces a few percent difference at 
              74 MHz.
    
    .. versionchanged:: 1.2.0
        Reworked the GSM model to use the actual GSM that has been 
        downsampled to 64 sides rather than the fit.
    """
    
    # Class data 
    degToRad = (pi/180.)             # Usual conversion factor
    
    _input = os.path.join(dataPath, 'skymap', 'gsm-408locked.npz')
    
    def __init__(self, skyMapFileName=None, freqMHz=73.9):
        """
        Initialize the SkyMapGSM object with an optional full file path to 
        the skymap file.
        """
        
        if skyMapFileName is None:
            skyMapFileName = self._input
            
        # Since we are using a pre-computed GSM which is a NPZ file, read it
        # in with numpy.load.
        dataDict = load(skyMapFileName)
        
        # RA and dec. are stored in the dictionary as radians
        self.ra = dataDict['ra'].ravel() / self.degToRad
        self.dec = dataDict['dec'].ravel() / self.degToRad
        
        # Compute the temperature for the current frequency
        ## Load in the data
        freqs = dataDict['freqs']
        sigmas = dataDict['sigmas']
        comps = dataDict['comps']
        maps = dataDict['maps']
        ## Build the scale and spectral component interpolation functions so that we can
        ## move from the surveys to an arbitrary frequency.  This is done using cubic 
        ## interpolation across log(freq)
        scaleFunc = interp1d(log(freqs), log(sigmas), kind='cubic')
        compFuncs = []
        for i in xrange(comps.shape[1]):
            compFuncs.append( interp1d(log(freqs), comps[:,i], kind='cubic') )
        ## Actually create the realization by running the interplation and adding up the
        ## compnent maps
        output = maps[:,0]*0.0
        for i,compFunc in enumerate(compFuncs):
            output += compFunc(log(freqMHz))*maps[:,i]
        output *= exp(scaleFunc(log(freqMHz)))
        ## Save
        self._power = output
        
        # Close out the dictionary
        try:
            dataDict.close()
        except AttributeError:
            pass
            
    def NormalizePower(self):
        """
        Compute the skymap power (total power radiated into 4 pi steradians) into 
        a power at antenna, based on pixel count.
        """
        
        return self._power
        
    def ComputeTotalPowerFromSky(self):
        """
        Compute and return the the total power from the sky.
        """
        
        if len(self._power) == 0:
            raise SkyMapPowerError("self._power contains 0 elements")
        return self.NormalizePower().sum()


class SkyMapLFSM(SkyMapGSM):
    """
    Extension of the SkyMap class to use the Low Frequency Sky Model with 5.1
    degree resolution.
    
    For more information on the Low Frequency Sky Model, see: https://lda10g.alliance.unm.edu/LWA1LowFrequencySkySurvey/
    
    .. versionadded:: 1.2.0
    """
    
    _input = os.path.join(dataPath, 'skymap', 'lfsm-5.1deg.npz')
    
    def __init__(self, skyMapFileName=None, freqMHz=73.9):
        """
        Initialize the SkyMapLFSM object with an optional full file path to 
        the skymap file.
        """
        
        if skyMapFileName is None:
            skyMapFileName = self._input
            
        # Since we are using a pre-computed GSM which is a NPZ file, read it
        # in with numpy.load.
        dataDict = load(skyMapFileName)
        
        # RA and dec. are stored in the dictionary as radians
        self.ra = dataDict['ra'].ravel() / self.degToRad
        self.dec = dataDict['dec'].ravel() / self.degToRad
        
        # Compute the temperature for the current frequency
        ## Load in the data
        freqs = dataDict['freqs']
        sigmas = dataDict['sigmas']
        comps = dataDict['comps']
        maps = dataDict['maps']
        ## Build the scale and spectral component interpolation functions so that we can
        ## move from the surveys to an arbitrary frequency.  This is done using spline
        ## interpolation for the scale factor and cubic for the 2-D structure.
        ## interpolation across log(freq)
        scaleFunc = interp1d(log(freqs), log(sigmas), kind='slinear')
        compFuncs = []
        for i in xrange(comps.shape[1]):
            compFuncs.append( interp1d(log(freqs), comps[:,i], kind='cubic') )
        ## Actually create the realization by running the interplation and adding up the
        ## compnent maps
        output = maps[:,0]*0.0
        for i,compFunc in enumerate(compFuncs):
            output += compFunc(log(freqMHz))*maps[:,i]
        output *= exp(scaleFunc(log(freqMHz)))
        ## Save
        self._power = output
        
        # Close out the dictionary
        try:
            dataDict.close()
        except AttributeError:
            pass


class ProjectedSkyMap(object):
    """
    The class for handling the model sky brightness maps over a particular site.
    This code is the base class for the sky map visible at a specific location. It 
    takes as input a skymap file name and frequency to which the skymap corresponds.
    It inherits from class SkyMap. It has the following methods:
    1. _init_ - takes the array coordinate filename as an input argument.
    2. ComputeDirectionCosines - Computes the direction cosines 
    3. ComputeTotalPowerFromVisibleSky - Sums the power for all visible sources in 
    the sky.
    """
    
    def __init__(self, skyMapObject, lat, lon, utc_jd):
        """
        Initialize the skymap at input lat,lon (decimal degrees) and time (in 
        UTC julian day).
        """
        
        self.skyMapObject = skyMapObject
        
        assert  -90 <= lat <=  90, ValueError('lat = %g not in [-90,90]' % lat)
        assert -360 <= lon <= 360, ValueError('lon = %g not in [-360,360]' % lon)
        self.lat  = lat
        self.lon  = lon
        self.time = utc_jd
        
        # Compute the ra and dec locations to a altitude and azimuth. This requires a lat/lon and a time (UTC).
        # Alt and Az are compressed to show only the visible source.
        ## Extract the RA/dec values
        ra = self.skyMapObject.ra
        dec = self.skyMapObject.dec
        
        ## Compute the LST at the lat/lon (hours)
        lst = astro.get_local_sidereal_time(self.lon, self.time)
        
        ## RA -> HA (deg)
        ha = lst*15.0 - ra
        
        ## HA, dec, and lat in degrees to alt and az in radians
        sinAlt = sin(dec*pi/180)*sin(self.lat*pi/180) + cos(dec*pi/180)*cos(lat*pi/180)*cos(ha*pi/180)
        sinAlt = clip(sinAlt, -1.0, 1.0)
        alt = arcsin(sinAlt)
        cosAz = (sin(dec*pi/180)-sinAlt*sin(self.lat*pi/180))/(cos(alt)*cos(lat*pi/180))
        cosAz = clip(cosAz, -1.0, 1.0)
        az = arccos(cosAz)
        swap = where(sin(ha*pi/180) > 0)
        az[swap] = 2*pi-az[swap]
        
        ## Convert to alt and az to degrees
        self.alt = alt*180/pi
        self.az  = az*180/pi
        
        # Compress the az/alt so that only the visible sources are available. "
        visibleMask = self.alt > 0.0
        
        #Compress arrays to hide non-visible sources
        self.visibleAlt = compress(visibleMask, self.alt)
        self.visibleAz = compress(visibleMask, self.az)
        self.visiblePower = compress(visibleMask, self.skyMapObject._power)
        try:
            pixelSolidAngle = (2.0 * pi)**2/(self.skyMapObject.numPixelsX*self.skyMapObject.numPixelsY)
        except AttributeError:
            pixelSolidAngle = 2.5566346e-4
        fractionSolidAngle = (1.0/4.0*pi) * pixelSolidAngle
        
        # The cosine term is the projection of the receiving area onto the direction 
        # of the source
        normalizedPower = self.skyMapObject.NormalizePower() * fractionSolidAngle  
        self.visibleNormalizedPower = compress(visibleMask, normalizedPower)
        #self.visibleNormalizedPower = self.visiblePower*cos(self.visibleAlt * self.skyMapObject.degToRad)
        self.visibleRa = compress(visibleMask, self.skyMapObject.ra)
        self.visibleDec = compress(visibleMask, self.skyMapObject.dec)
        
    def ComputeDirectionCosines(self):
        """
        Compute the direction cosines and return the tuple of arrays (l,m,n).
        """
        
        altRad = self.visibleAlt*self.skyMapObject.degToRad
        azRad = self.visibleAz*self.skyMapObject.degToRad
        l = cos(altRad)*sin(azRad)
        m = cos(altRad)*cos(azRad)
        n = sin(altRad)
        return (l, m, n)
        
    def ComputeTotalPowerFromVisibleSky(self):
        """
        Compute and return the the total power from visible portion of the sky.
        """
        
        if len(self.visibleNormalizedPower) == 0:
            raise SkyMapPowerError("visibleNormalizedPower contains 0 elements")
        totalVisiblePower = sum(self.visibleNormalizedPower)
        return totalVisiblePower



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Module for generating simulated arrays and visibility data.  The chief 
functions of this module are:

buildSimArray
  given a station object, a list of stands, and a list of frequencies, build 
  a AIPY AntennaArray-like object.  This module can also generate AntennaArray 
  objects with positional errors by setting the 'PosError' keyword to a 
  positive value.

buildSimData
  given a SimArray and a list of aipy.src sources, build up a collection of 
  visibilities for a given set of Julian dates

scaleData
  given a dictionary of simulated visibilities from buildSimData, apply 
  antenna-based gains and delays to the visibilities

shiftData
  given a dictionary of simulated visibilities from buildSimData, shift the uvw 
  coordinates of the visibilities.
  .. note::
    This only changes the uvw values and does not phase-shift the data.

The format of the data dictionaries mentioned above is:

primary keys
  The primary keys store the major aspects of the visiblity data, e.g., 
  frequency coverage, baseline pairs, uvw coordinates, etc.  Valid keys are:
    * *freq* - list of frequencies used in Hz
    * *isMasked* - whether or not the visibility data have been masked 
    (numpy.compress'd)
    * *bls* - list of baselines in (stand 1, stand2) format
    * *uvw* - list of uvw coordinates as 3-element numpy arrays
    * *vis* - list of visibility numpy arrays
    * *wgt* - list of weight arrays with the same length as the visilitity arrays
    * *msk* - list of mask arrays used for the data.  1 = masked, 0 = valid
    * *jd*  - list of Julian dates associated with each list element

secondary keys
  The bls, uvw, vis, wgt, msk, and jd primary keys also have secondary keys that 
  indicate which polarizations are being stored.  Valid keys are:
    * *xx*
    * *yy*
    * *xy*
    * *yx*

In addition to simulation functions, this module includes buildGriddedImage
which takes a dictionary of visibilities and returns and aipy.im.ImgW object.

.. versionchanged:: 0.3.0
    This module was formerly called lsl.sim.sim
    
.. versionchanged:: 0.5.0
    Moved buildGriddedImage to the :mod:`lsl.imaging.utils` module.

.. versionchanged:: 1.0.1
    Switched over to a new C-based simulation package
    
.. versionchanged:: 1.0.2
    Added in a function, addBaselineNoise, to help bring noise into the 
    simulations
    
.. versionchanged:: 1.0.3
    Moved the calculateSEFD function into lsl.misc.rfutils
    Changed the meaning of the ForceGaussian parameter of the buildSimArray()
    function to be the Gaussian full width at half maximum in degrees
"""

import os
import aipy
import copy
import math
import ephem
import numpy
import warnings
from scipy.interpolate import interp1d

from lsl import astro
from lsl.common.paths import data as dataPath
from lsl.correlator import uvUtils
from lsl.common.stations import lwa1
from lsl.common.constants import c as speedOfLight

from _simfast import FastVis

__version__ = '0.6'
__revision__ = '$Rev: 2505 $'
__all__ = ['srcs', 'RadioEarthSatellite', 'BeamAlm', 'Antenna', 'AntennaArray', 
        'buildSimArray', 'buildSimData', 'scaleData', 'shiftData', 'addBaselineNoise', 
        '__version__', '__revision__', '__all__']


# A dictionary of bright sources in the sky to use for simulations
srcs = aipy.src.get_catalog(srcs=['Sun', 'Jupiter', 'cas', 'crab', 'cyg', 'her', 'sgr', 'vir'])


class RadioEarthSatellite(object):
    """
    Implement a aipy.amp.RadioBody-lime simulation object for an Earth-
    orbiting satellite using a two-line element set.
    """
    
    def __init__(self, tle, tfreq, tpower=0.0, tbw=1.0e6, ionref=(0.,0.)):
        """
        Initialize the class using:
          * a three-element list of strings that specify the two-line element
            set of the satellite,
          * a list of frequencies (in GHz) where the satellite transmits, 
          * the transmitter power in W, and
          * the transmission bandwidth in Hz.
        
        .. note::
            For calculating the power received on the ground this class 
            assumes that the signal is emitted isotropically.
        """
        
        # Location
        self.Body = ephem.readtle(*tle)
        self.src_name = self.Body.name
        
        # Transmitter - tuning, power, and bandwidth
        try:
            self.tfreq = list(tfreq)
        except TypeError:
            self.tfreq = [tfreq,]
        self.tpower = tpower
        self.tbw = tbw
        
        # Ionospheric distortion
        self.ionref = list(ionref)
        
        # Shape (unresolved)
        self.srcshape = list((0.,0.,0.))
        
    def __getattr__(self, nm):
        """
        First try to access attribute from this class, but if that fails, 
        try to get it from the underlying PyEphem object.
        """
        
        try:
            return object.__getattr__(self, nm)
        except AttributeError:
            return self.Body.__getattribute__(nm)
            
    def compute(self, observer):
        """
        Update coordinates relative to the provided observer.  Must be
        called at each time step before accessing information.
        """
        
        self.Body.compute(observer)
        self.update_jys(observer.get_afreqs())
        
    def get_crds(self, crdsys, ncrd=3):
        """
        Return the coordinates of this location in the desired coordinate
        system ('eq','top') in the current epoch.  If ncrd=2, angular
        coordinates (ra/dec or az/alt) are returned, and if ncrd=3,
        xyz coordinates are returned.
        """
        
        assert(crdsys in ('eq','top'))
        assert(ncrd in (2,3))
        if crdsys == 'eq':
            if ncrd == 2:
                return (self.ra, self.dec)
            return aipy.coord.radec2eq((self.ra, self.dec))
        else:
            if ncrd == 2:
                return (self.az, self.alt)
            return aipy.coord.azalt2top((self.az, self.alt))
            
    def update_jys(self, afreqs):
        """
        Update fluxes relative to the provided observer.  Must be
        called at each time step before accessing information.
        """
        
        # Setup
        r = self.Body.range				# m
        v = self.Body.range_velocity		# m/s
        self.jys = numpy.zeros_like(afreqs)
        
        # Compute the flux coming from the satellite assuming isotropic emission
        self._jys = self.tpower / (4*numpy.pi*r**2) / self.tbw		# W / m^2 / Hz
        self._jys /= 10**-26								# Jy
        
        try:
            dFreq = afreqs[1]-afreqs[0]
        except IndexError:
            dFreq = afreqs[0,1]-afreqs[0,0]
        for f in self.tfreq:
            ## Apply the Doppler shift
            fPrime = f / (1 + v/speedOfLight)
            
            ## Figure out which frequency bin is within one channel of the
            ## shifted signal.  If it's close, set it to a 
            diff = numpy.abs(fPrime - afreqs)*1e9
            good = numpy.where( diff <= self.tbw/2.0 )
            self.jys[good] = self._jys * min([1.0, dFreq*1e9/self.tbw])
                
    def get_jys(self):
        """
        Return the fluxes vs. freq that should be used for simulation.
        """
        
        return self.jys


class BeamAlm(aipy.amp.BeamAlm):
    """
    AIPY-based representation of a beam model where each pointing has a 
    response defined as a polynomial in frequency, and the spatial 
    distributions of these coefficients decomposed into spherical 
    harmonics.
    
    This differs from the AIPY version in that the response() method 
    accepts two and three-dimensions arrays of topocentric coordinates, 
    similar to what aipy.img.ImgW.get_top() produces, and computes the 
    beam response at all points.
    """
    
    def __init__(self, freqs, lmax=8, mmax=8, deg=7, nside=64, coeffs={}):
        """
        AIPY __init__() function.
        
        lmax = maximum spherical harmonic term
        mmax = maximum spherical harmonic term in the z direction
        deg = order of polynomial to used for mapping response of each pointing
        nside = resolution of underlying HealpixMap to use
        coeffs = dictionary of polynomial term (integer) and corresponding Alm 
        coefficients (see healpix.py doc).
        """
        
        aipy.phs.Beam.__init__(self, freqs)
        self.alm = [aipy.healpix.Alm(lmax, mmax) for i in range(deg+1)]
        self.hmap = [aipy.healpix.HealpixMap(nside, scheme='RING', interp=True)
            for a in self.alm]
        for c in coeffs:
            if c < len(self.alm):
                self.alm[-1-c].set_data(coeffs[c])
        self._update_hmap()
        
    def __responsePrimitive(self, top):
        """
        Copy of the original aipy.amp.BeamAlm.response function.
        
        Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y').
        """
        
        top = [aipy.healpix.mk_arr(c, dtype=numpy.double) for c in top]
        px,wgts = self.hmap[0].crd2px(*top, interpolate=1)
        poly = numpy.array([numpy.sum(h.map[px] * wgts, axis=-1) for h in self.hmap])
        rv = numpy.polyval(poly, numpy.reshape(self.afreqs, (self.afreqs.size, 1)))
        return rv
        
    def response(self, top):
        """
        Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y').
        
        .. note::
            This function also accepts two and three-dimensions arrays of 
            topocentric coordinates, similar to what aipy.img.ImgW.get_top() 
            produces, and computes the beam response at all points
        """
        
        test = numpy.array(top)
        x,y,z = top
        
        if len(test.shape) == 1:
            temp = self.__responsePrimitive((x,y,z))
            
        elif len(test.shape) == 2:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                temp[:,i] = numpy.squeeze(self.__responsePrimitive((x[i],y[i],z[i])))
                
        elif len(test.shape) == 3:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                for j in xrange(temp.shape[2]):
                    temp[:,i,j] = numpy.squeeze(self.__responsePrimitive((x[i,j],y[i,j],z[i,j])))
                    
        else:
            raise ValueError("Cannot compute response for %s" % str(test.shape))
            
        return temp


class Beam2DGaussian(aipy.amp.Beam2DGaussian):
    """
    AIPY-based representation of a 2-D Gaussian beam pattern, with default 
    setting for a flat beam.  The width parameters denote the 'sigma' of
    the Gaussian in radians.
    
    This differs from the AIPY version in that the response() method 
    accepts two and three-dimensions arrays of topocentric coordinates, 
    similar to what aipy.img.ImgW.get_top() produces, and computes the 
    beam response at all points.
    
    .. versionchanged:: 1.0.3
        Clarified what 'xwidth' and 'ywidth' are.
    """
    
    def __init__(self, freqs, xwidth=numpy.Inf, ywidth=numpy.Inf):
        """
        AIPY __init__() function.
        
        xwidth = angular width (radians) in EW direction
        ywidth = angular width (radians) in NS direction
        """
        
        aipy.phs.Beam.__init__(self, freqs)
        self.xwidth, self.ywidth = xwidth, ywidth
        
    def __responsePrimitive(self, top):
        """
        Copy of the original aipy.amp.Beam2DGaussian.response function.
        
        Return beam response across active band for specified topocentric 
        coordinates: (x=E,y=N,z=UP). x,y,z may be arrays of multiple 
        coordinates.  Returns 'x' linear polarization (rotate pi/2 for 'y').
        """
        
        x,y,z = top
        x,y = numpy.arcsin(x)/self.xwidth, numpy.arcsin(y)/self.ywidth
        resp = numpy.sqrt(numpy.exp(-(x**2 + y**2)/2.0))
        resp = numpy.resize(resp, (self.afreqs.size, resp.size))
        return resp
        
    def response(self, top):
        """
        Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y').
        
        .. note::
            This function also accepts two and three-dimensions arrays of 
            topocentric coordinates, similar to what aipy.img.ImgW.get_top() 
            produces, and computes the beam response at all points
        """
        
        test = numpy.array(top)
        x,y,z = top
        
        if len(test.shape) == 1:
            temp = self.__responsePrimitive((x,y,z))
            
        elif len(test.shape) == 2:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                temp[:,i] = numpy.squeeze(self.__responsePrimitive((x[i],y[i],z[i])))
                
        elif len(test.shape) == 3:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                for j in xrange(temp.shape[2]):
                    temp[:,i,j] = numpy.squeeze(self.__responsePrimitive((x[i,j],y[i,j],z[i,j])))
                    
        else:
            raise ValueError("Cannot compute response for %s" % str(test.shape))
            
        return temp


class BeamPolynomial(aipy.amp.BeamPolynomial):
    """
    AIPY-based representation of a Gaussian beam model whose width varies 
    with azimuth angle and with frequency.
    
    This differs from the AIPY version in that the response() method 
    accepts two and three-dimensions arrays of topocentric coordinates, 
    similar to what aipy.img.ImgW.get_top() produces, and computes the 
    beam response at all points.
    """
    def __init__(self, freqs, poly_azfreq=numpy.array([[.5]])):
        """
        AIPY __init__() function.
        
        poly_azfreq = a 2D polynomial in cos(2*n*az) for first axis and 
        in freq**n for second axis.
        """
        
        self.poly = poly_azfreq
        aipy.phs.Beam.__init__(self, freqs)
        self.poly = poly_azfreq
        self._update_sigma()
        
    def __responsePrimitive(self, top):
        """
        Copy of the original aipy.amp.Beam2DGaussian.response function.
        
        Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y').
        """
        
        az,alt = aipy.coord.top2azalt(top)
        zang = numpy.pi/2 - alt
        if zang.size == 1:
            zang = numpy.array([zang]); zang.shape = (1,)
            az = numpy.array([az])
            az.shape = (1,)
        a = 2 * numpy.arange(self.poly.shape[0], dtype=numpy.float)
        a.shape = (1,) + a.shape
        az.shape += (1,)
        zang.shape += (1,)
        a = numpy.cos(numpy.dot(az, a))
        a[:,0] = 0.5
        s = numpy.dot(a, self.sigma)
        return numpy.sqrt(numpy.exp(-(zang/s)**2)).transpose()
        
    def response(self, top):
        """
        Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y').
        
        .. note::
            This function also accepts two and three-dimensions arrays of 
            topocentric coordinates, similar to what aipy.img.ImgW.get_top() 
            produces, and computes the beam response at all points
        """
        
        test = numpy.array(top)
        x,y,z = top
        
        if len(test.shape) == 1:
            temp = self.__responsePrimitive((x,y,z))
            
        elif len(test.shape) == 2:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                temp[:,i] = numpy.squeeze(self.__responsePrimitive((x[i],y[i],z[i])))
                
        elif len(test.shape) == 3:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                for j in xrange(temp.shape[2]):
                    temp[:,i,j] = numpy.squeeze(self.__responsePrimitive((x[i,j],y[i,j],z[i,j])))
                    
        else:
            raise ValueError("Cannot compute response for %s" % str(test.shape))
            
        return temp


class Beam(aipy.amp.Beam):
    """
    AIPY-based representation of a flat (gain=1) antenna beam pattern.
    
    This differs from the AIPY version in that the response() method 
    accepts two and three-dimensions arrays of topocentric coordinates, 
    similar to what aipy.img.ImgW.get_top() produces, and computes the 
    beam response at all points.
    """
    
    def __responsePrimitive(self, top):
        """
        Copy of the original aipy.amp.Beam.response function.
        
        Return the (unity) beam response as a function of position.
        """
        
        x,y,z = numpy.array(top)
        return numpy.ones((self.afreqs.size, x.size))
        
    def response(self, top):
        """
        Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y').
        
        .. note::
            This function also accepts two and three-dimensions arrays of 
            topocentric coordinates, similar to what aipy.img.ImgW.get_top() 
            produces, and computes the beam response at all points
        """
        
        test = numpy.array(top)
        x,y,z = top
        
        if len(test.shape) == 1:
            temp = self.__responsePrimitive((x,y,z))
            
        elif len(test.shape) == 2:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                temp[:,i] = numpy.squeeze(self.__responsePrimitive((x[i],y[i],z[i])))
                
        elif len(test.shape) == 3:
            temp = numpy.zeros((self.afreqs.size,)+test.shape[1:])
            for i in xrange(temp.shape[1]):
                for j in xrange(temp.shape[2]):
                    temp[:,i,j] = numpy.squeeze(self.__responsePrimitive((x[i,j],y[i,j],z[i,j])))
                    
        else:
            raise ValueError("Cannot compute response for %s" % str(test.shape))
            
        return temp


class Antenna(aipy.amp.Antenna):
    """
    Modification to the aipy.amp.Antenna class to also store the stand ID 
    number in the Antenna.stand attribute.  This also add a getBeamShape 
    attribute that pulls in the old vis.getBeamShape function.
    """

    def __init__(self, x, y, z, beam, phsoff=[0.,0.], bp_r=numpy.array([1]), bp_i=numpy.array([0]), amp=1, pointing=(0.,numpy.pi/2,0), stand=0, **kwargs):
        """
        New init function that include stand ID number support.  From aipy.amp.Antenna:
          * x,y z = antenna coordinates in equatorial (ns) coordinates
          * beam = Beam object (implements response() function)
          * phsoff = polynomial phase vs. frequency.  Phs term that is linear
                     with freq is often called 'delay'.
          * bp_r = polynomial (in freq) modeling real component of passband
          * bp_i = polynomial (in freq) modeling imaginary component of passband
          * amp = overall multiplicative scaling of gain
          * pointing = antenna pointing (az,alt).  Default is zenith.
        """
        
        aipy.phs.Antenna.__init__(self, x,y,z, beam=beam, phsoff=phsoff)
        self.set_pointing(*pointing)
        self.bp_r = bp_r
        self.bp_i = bp_i
        self.amp = amp
        self.stand = stand
        self._update_gain()
        
    def bm_response(self, top, pol='x'):
        """
        Return response of beam for specified polarization.
        
        .. note::
            This differs from the AIPY implementation in that the LWA X-pol.
            is oriented N-S, not E-W.
            
        .. note::
            This function also accepts two and three-dimensions arrays of 
            topocentric coordinates, similar to what img.ImgW.get_top() 
            produces, and computes the beam response at all points.
        """
        top = numpy.array(top)
        
        def robustDot(a, b):
            """
            Dot product that operations on multi-dimensional coordinate sets.
            """
            
            if len(b.shape) == 1:
                temp = numpy.dot(a, b)
                
            elif len(b.shape) == 2:
                temp = numpy.zeros_like(b)
                for i in xrange(b.shape[1]):
                    temp[:,i] = numpy.dot(a, b[:,i])
                    
            elif len(b.shape) == 3:
                temp = numpy.zeros_like(b)
                for i in xrange(b.shape[1]):
                    for j in xrange(top.shape[2]):
                        temp[:,i,j] = numpy.dot(a, b[:,i,j])
                        
            else:
                raise ValueError("Cannot dot a (%s) with b (%s)" % (str(a.shape), str(b.shape)))
            
            return temp
            
        top = {'y':robustDot(self.rot_pol_x, top), 
            'x':robustDot(self.rot_pol_y, top), 
            'l':robustDot(self.rot_pol_x, top), 
            'r':robustDot(self.rot_pol_y, top)}[pol]
        x,y,z = top
        
        return self.beam.response((x,y,z))
        
    def get_beam_shape(self, pol='x'):
        """
        Return a 360 by 90 by nFreqs numpy array showing the beam pattern of a
        particular antenna in the array.  The first two dimensions of the output 
        array contain the azimuth (from 0 to 359 degrees in 1 degree steps) and 
        altitude (from 0 to 89 degrees in 1 degree steps).
        """
        
        # Build azimuth and altitude arrays.  Be sure to convert to radians
        az = numpy.zeros((360,90))
        for i in range(360):
            az[i,:] = i*numpy.pi/180.0
        alt = numpy.zeros((360,90))
        for i in range(90):
            alt[:,i] = i*numpy.pi/180.0
            
        # The beam model is computed in terms of topocentric coordinates, so make that
        # conversion right quick using the aipy.coord module.
        xyz = aipy.coord.azalt2top(numpy.concatenate([[az],[alt]]))
        
        # I cannot figure out how to do this all at once, so loop through azimuth/
        # altitude pairs
        resp = numpy.zeros((360,90,len(self.beam.freqs)))
        for i in range(360):
            for j in range(90):
                resp[i,j,:] = numpy.squeeze( self.bm_response(numpy.squeeze(xyz[:,i,j]), pol=pol) )
                
        return resp


class AntennaArray(aipy.amp.AntennaArray):
    """
    Modification to the aipy.ant.AntennaArray class to add a function to 
    retrieve the stands stored in the AntennaArray.ants attribute.  Also add 
    a function to set the array time from a UNIX timestamp.
    
    .. versionchanged:: 1.0.1
        Added an option to set the ASP filter for simulation proposes.  
        This updates the bandpasses used by AIPY to include the antenna
        impedance mis-match and the mean ARX response.
    """

    def get_stands(self):
        """
        Return a numpy array listing the stands found in the AntennaArray 
        object.
        """

        stands = []
        for ant in self.ants:
            stands.append(ant.stand)
        
        return numpy.array(stands)

    def set_unixtime(self, timestamp):
        """
        Set the array time using a UNIX timestamp (epoch 1970).
        """
        
        self.set_jultime(astro.unix_to_utcjd(timestamp))
        
    def set_asp_filter(self, filter='split'):
        """
        Update the bandpasses for the antennas to include the effect of 
        the antenna impedance mis-match (IMM) and the mean LWA1 ARX 
        response.
        
        Valid filters are:
          * split
          * full
          * reduced
          * split@3MHz
          * full@3MHz
          * none
        
        None is a special case where both the IMM and ARX response are 
        removed, i.e., the bandpass is unity for all frequencies.
        
        .. versionchanged:: 1.2.1
            Added support for the 'split@3MHz' and 'full@3MHz' filters at
            LWA-SV.
        
        .. versionadded:: 1.0.1
        """
        
        # Get the frequencies we need to estimate the bandpass for
        freqs = self.get_afreqs()*1e9
        if len(freqs.shape) == 2:
            freqs = freqs[0,:]
            
        if filter == 'none' or filter is None:
            # Build up the bandpass - of ones
            resp = numpy.ones(freqs.size)
            
        else:
            # Load in the LWA1 antennas so we can grab some data
            ants = lwa1.getAntennas()
            
            # Antenna impedance mis-match
            immf, immr = ants[0].response(dB=False)
            immr /= immr.max()
            immIntp = interp1d(immf, immr, kind='cubic', bounds_error=False)
            
            # Mean ARX filter response
            arxf, arxr = ants[0].arx.response(dB=False)
            for i in xrange(1, len(ants)):
                arxfi, arxri = ants[i].arx.response(dB=False)
                arxr += arxri
            arxr /= arxr.max()
            arxIntp = interp1d(arxf, arxr, kind='cubic', bounds_error=False)
            
            # Build up the bandpass
            resp = numpy.ones(freqs.size)
            resp *= immIntp(freqs)
            resp *= arxIntp(freqs)
            
        # Update the AIPY passbands
        for i in xrange(len(self.ants)):
            self[i].amp = resp
            self[i].update()
            
    def get_baseline_fast(self, i, j, src='z', map=None):
        """Return the baseline corresponding to i,j in various coordinate 
        projections: src='e' for current equatorial, 'z' for zenith 
        topocentric, 'r' for unrotated equatorial, or a RadioBody for
        projection toward that source - fast."""
        bl = self[j] - self[i]
        
        if type(src) == str:
            if src == 'e':
                return numpy.dot(self._eq2now, bl)
            elif src == 'z':
                return numpy.dot(self._eq2zen, bl)
            elif src == 'r':
                return bl
            else:
                raise ValueError('Unrecognized source:' + src)
        try:
            if src.alt < 0:
                raise PointingError('%s below horizon' % src.src_name)
            m = src.map
        except AttributeError:
            if map is None:
                ra,dec = aipy.coord.eq2radec(src)
                m = aipy.coord.eq2top_m(self.sidereal_time() - ra, dec)
            else:
                m = map
        return numpy.dot(m, bl).transpose()
        
    def gen_uvw_fast(self, i, j, src='z', w_only=False, map=None):
        """Compute uvw coordinates of baseline relative to provided RadioBody, 
        or 'z' for zenith uvw coordinates.  If w_only is True, only w (instead
        of (u,v,w) will be returned) - fast."""
        
        x,y,z = self.get_baseline_fast(i,j, src=src, map=map)
        
        afreqs = self.get_afreqs()
        afreqs = numpy.reshape(afreqs, (1,afreqs.size))
        if len(x.shape) == 0:
            if w_only:
                return z*afreqs
            else:
                return numpy.array([x*afreqs, y*afreqs, z*afreqs])
                
        #afreqs = numpy.reshape(afreqs, (1,afreqs.size))
        x.shape += (1,); y.shape += (1,); z.shape += (1,)
        
        if w_only:
            out = numpy.dot(z,afreqs)
        else:
            out = numpy.array([numpy.dot(x,afreqs), numpy.dot(y,afreqs), numpy.dot(z,afreqs)])
            
        return out
        
    def gen_phs_fast(self, src, i, j, mfreq=.150, ionref=None, srcshape=None, resolve_src=False, u=None, v=None, w=None):
        """Return phasing that is multiplied to data to point to src - fast."""
        if ionref is None:
            try:
                ionref = src.ionref
            except AttributeError:
                pass
        if ionref is not None or resolve_src:
            if u is None or v is None or w is None:
                u,v,w = self.gen_uvw_fast(i,j,src=src)
        else:
            if w is None:
                w = self.gen_uvw_fast(i,j,src=src, w_only=True)
        if ionref is not None:
            w += self.refract(u, v, mfreq=mfreq, ionref=ionref)
        o = self.get_phs_offset(i,j)
        phs = numpy.exp(-1j*2*numpy.pi*(w + o))
        if resolve_src:
            if srcshape is None:
                try:
                    srcshape = src.srcshape
                except AttributeError:
                    pass
        if srcshape is not None:
            phs *= self.resolve_src(u, v, srcshape=srcshape)
        return phs.squeeze()
        
    def sim(self, i, j, pol='xx'):
        """
        Simulate visibilities for the specified (i,j) baseline and 
        polarization.  sim_cache() must be called at each time step before 
        this will return valid results.
        
        This function differs from aipy.amp.AntennaArray.sim in the fact that
        *ionref* and *srcshape* are both None in the call to gen_phs and that
        *resolve_src* is set to False.
        """
        
        assert(pol in ('xx','yy','xy','yx'))
        
        if self._cache is None:
            raise RuntimeError('sim_cache() must be called before the first sim() call at each time step.')
        elif self._cache == {}:
            return numpy.zeros_like(self.passband(i,j))
            
        s_eqs = self._cache['s_eqs']
        s_map = self._cache['s_map']
        w = self.gen_uvw_fast(i, j, src=s_eqs, map=s_map, w_only=True)
        I_sf = self._cache['jys']
        Gij_sf = self.passband(i,j)
        Bij_sf = self.bm_response(i, j, pol=pol)
        if len(Bij_sf.shape) == 2:
            Gij_sf = numpy.reshape(Gij_sf, (1, Gij_sf.size))
            
        # Get the phase of each src vs. freq, also does resolution effects
        E_sf = numpy.conjugate( self.gen_phs_fast(s_eqs, i, j, mfreq=self._cache['mfreq'], resolve_src=False, w=w) )
        try:
            E_sf.shape = I_sf.shape
        except(AttributeError):
            pass
            
        # Combine and sum over sources
        GBIE_sf = Gij_sf * Bij_sf * I_sf * E_sf
        Vij_f = GBIE_sf.sum(axis=0)
        
        return Vij_f


def buildSimArray(station, antennas, freq, jd=None, PosError=0.0, ForceFlat=False, ForceGaussian=False, verbose=False):
    """
    Build a AIPY AntennaArray for simulation purposes.  Inputs are a station 
    object defined from the lwa_common module, a numpy array of stand 
    numbers, and a numpy array of frequencies in either Hz of GHz.  Optional 
    inputs are a Julian Date to set the array to and a positional error terms 
    that perturbs each of the stands in x, y, and z.  The output of this 
    module is an AIPY AntennaArray object.
    
    The shape of the antenna response is either flat (gain of 1 in all 
    directions), modeled by a 2-D Gaussian with the specified full width at
    half maximum in degrees, or modeled by a collection of spherical 
    harmonics that are polynomials in frequency.  The spherical harmonics 
    are used if the file 'beam_shape.npz' is found in the current directory.
    
    .. versionchanged:: 1.0.3
        Changed the meaning of the ForceGaussian parameters so that the
        Gaussian full width at half maximum in degrees is passed in.
        
    .. versionchanged:: 1.0.1
        Moved the simulation code over from AIPY to the new _simFast module.  
        This should be much faster but under the caveats that the bandpass
        and antenna gain patterns are the same for all antennas.  This 
        should be a reasonable assumption for large-N arrays.
        
        Added an option to use a 2-D Gaussian beam pattern via the ForceGaussian
        keyword.
        
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.station` module instead of a list of stand ID numbers.
    """
    
    # If the frequencies are in Hz, we need to convert to GHz
    freqs = freq.copy()
    if freqs.min() > 1e6:
        freqs /= 1.0e9
        
    # If the beam Alm coefficient file is present, build a more realistic beam 
    # response.  Otherwise, assume a flat beam
    if ForceGaussian:
        try:
            xw, yw = ForceGaussian
            xw, yw = float(xw), float(yw)
        except (TypeError, ValueError) as e:
            xw = float(ForceGaussian)
            yw = 1.0*xw
            
        # FWHM to sigma
        xw /= 2.0*numpy.sqrt(2.0*numpy.log(2.0))
        yw /= 2.0*numpy.sqrt(2.0*numpy.log(2.0))
        
        # Degrees to radians
        xw *= numpy.pi/180
        yw *= numpy.pi/180
        
        if verbose:
            print("Using a 2-D Gaussian beam with sigmas %.1f by %.1f degrees" % (xw*180/numpy.pi, yw*180/numpy.pi))
        beam = Beam2DGaussian(freqs, xw, yw)
        
    elif ForceFlat:
        if verbose:
            print("Using flat beam model")
        beam = Beam(freqs)
        
    else:
        if os.path.exists(os.path.join(dataPath, 'beam-shape.npz')):
            dd = numpy.load(os.path.join(dataPath, 'beam-shape.npz'))
            coeffs = dd['coeffs']
            
            deg = coeffs.shape[0]-1
            lmax = int((math.sqrt(1+8*coeffs.shape[1])-3)/2)
            beamShapeDict = {}
            for i in range(deg+1):
                beamShapeDict[i] = numpy.squeeze(coeffs[-1-i,:])
            try:
                dd.close()
            except AttributeError:
                pass
                
            if verbose:
                print("Using Alm beam model with %i-order freq. polynomial and %i-order sph. harmonics" % (deg, lmax))
            beam = BeamAlm(freqs, lmax=lmax, mmax=lmax, deg=deg, nside=128, coeffs=beamShapeDict)
        else:
            if verbose:
                print("Using flat beam model")
            beam = Beam(freqs)
            
    if PosError != 0:
        warnings.warn("Creating array with positional errors between %.3f and %.3f m" % (-PosError, PosError), RuntimeWarning)

    # Build an array of AIPY Antenna objects
    ants = []
    for antenna in antennas:
        top = numpy.array([antenna.stand.x, antenna.stand.y, antenna.stand.z])
        top += (2*PosError*numpy.random.rand(3)-PosError)	# apply a random positional error if needed
        top.shape = (3,)
        eq = numpy.dot( aipy.coord.top2eq_m(0.0, station.lat), top )
        eq *= 100	# m -> cm
        eq /= aipy.const.c	# cm -> s
        eq *= 1e9	# s -> ns
        
        delayCoeff = numpy.zeros(2)
        
        amp = 0*antenna.cable.gain(freq*1e9) + 1
        
        ants.append( Antenna(eq[0], eq[1], eq[2], beam, phsoff=delayCoeff, amp=amp, stand=antenna.stand.id) )

    # Combine the array of antennas with the array's location to generate an
    # AIPY AntennaArray object
    simAA = AntennaArray(ants=ants, location=station.getAIPYLocation())
    
    # Set the Julian Data for the AntennaArray object if it is provided.  The try...except
    # clause is used to deal with people who may want to pass an array of JDs in rather than
    # just one.  If one isn't provided, use the date set for the input 'station'.
    if jd is not None:
        try:
            junk = len(jd)
        except:
            simAA.set_jultime(jd)
        else:
            simAA.set_jultime(jd[0])
    else:
        simAA.date = station.date
        
    return simAA


def __buildSimData(aa, srcs, pols=['xx', 'yy', 'xy', 'yx'], jd=None, chan=None, phaseCenter='z', baselines=None, mask=None, verbose=False, count=None, max=None, flatResponse=False, resolve_src=False):
    """
    Helper function for buildSimData so that buildSimData can be called with 
    a list of Julian Dates and reconstruct the data appropriately.
    
    .. versionchanged:: 1.0.1
        * Moved the simulation code over from AIPY to the new _simFast module.  
          This should be much faster but under the caveats that the bandpass
          and antenna gain patterns are the same for all antennas.  This 
          should be a reasonable assumption for large-N arrays.
        * Added a 'flatResponse' keyword to make it easy to toggle on and off
          the spectral and spatial response of the array for the simulation
        * Added a 'resolve_src' keyword to turn on source resolution effects
    """
    
    rawFreqs = aa.get_afreqs()
    
    if len(rawFreqs.shape) == 2:
        nFreq = rawFreqs.shape[1]
    else:
        nFreq = rawFreqs.size
    if chan is None:
        chanMin = 0
        chanMax = -1
    else:
        chanMin = chan[0]
        chanMax = chan[-1]
        
    # Update the JD if necessary
    if jd is not None:
        if verbose:
            if count is not None and max is not None:
                print("Setting Julian Date to %.5f (%i of %i)" % (jd, count, max))
            else:
                print("Setting Julian Date to %.5f" % jd)
        aa.set_jultime(jd)
    else:
        jd = aa.get_jultime()
    Gij_sf = aa.passband(0,1)
    def Bij_sf(xyz, pol):
        Bi = aa[0].bm_response(xyz, pol=pol[0]).transpose()
        Bj = aa[1].bm_response(xyz, pol=pol[1]).transpose()
        Bij = numpy.sqrt( Bi*Bj.conj() )
        return Bij.squeeze()
    if flatResponse:
        Gij_sf *= 0.0
        Gij_sf += 1.0
        
    # Compute the source parameters
    srcs_tp = []
    srcs_eq = []
    srcs_ha = []
    srcs_dc = []
    srcs_jy = []
    srcs_fq = []
    srcs_sh = []
    if verbose:
        print("Sources Used for Simulation:")
    for name in srcs:
        ## Update the source's coordinates
        src = srcs[name]
        src.compute(aa)
        
        ## Remove sources below the horizon
        srcTop = src.get_crds(crdsys='top', ncrd=3)
        srcAzAlt = aipy.coord.top2azalt(srcTop)
        if srcAzAlt[1] < 0:
            if verbose:
                print("  %s: below horizon" % name)
            continue
            
        ## Topocentric coordinates for the gain pattern calculations
        srcTop.shape = (1,3)
        srcs_tp.append( srcTop )
        
        ## RA/dec -> equatorial 
        srcEQ = src.get_crds(crdsys='eq', ncrd=3)
        srcEQ.shape = (3,1)
        srcs_eq.append( srcEQ )
        
        ## HA/dec
        srcRA,srcDec = aipy.coord.eq2radec(srcEQ)
        srcHA = aa.sidereal_time() - srcRA
        srcs_ha.append( srcHA )
        srcs_dc.append( srcDec )
        
        ## Source flux over the bandpass - corrected for the bandpass
        jys = src.get_jys()
        jys.shape = (1,nFreq)
        jys = Gij_sf * jys
        srcs_jy.append( jys )
        
        ## Frequencies that the fluxes correspond to
        frq = aa.get_afreqs()
        frq.shape = (1,nFreq)
        srcs_fq.append( frq )
        
        ## Source shape parameters
        shp = numpy.array(src.srcshape)
        shp.shape = (3,1)
        srcs_sh.append( shp )
        
    # Build the simulation cache
    aa.sim_cache( numpy.concatenate(srcs_eq, axis=1), 
                jys=numpy.concatenate(srcs_jy, axis=0),
                mfreqs=numpy.concatenate(srcs_fq, axis=0),
                srcshapes=numpy.concatenate(srcs_sh, axis=1) )
    aa._cache['s_top'] = numpy.concatenate(srcs_tp, axis=0)
    aa._cache['s_ha'] = numpy.concatenate(srcs_ha, axis=0)
    aa._cache['s_dec'] = numpy.concatenate(srcs_dc, axis=0)
    
    # Build the simulated data.  If no baseline list is provided, build all 
    # baselines available
    if baselines is None:
        baselines = uvUtils.getBaselines(numpy.zeros(len(aa.ants)), Indicies=True)
        
    # Define output data structure
    freq = aa.get_afreqs()*1e9
    if len(freq.shape) == 2:
        freq = freq[0,:]
    UVData = {'freq': freq, 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
    if mask is None:
        UVData['isMasked'] = False
    else:
        UVData['isMasked'] = True
        
    # Go!
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcAz = phaseCenter.az*180/numpy.pi
        pcEl = phaseCenter.alt*180/numpy.pi
    else:
        pcAz = 0.0
        pcEl = 90.0
        
    for p,pol in enumerate(pols):
        ## Apply the antenna gain pattern for each source
        if not flatResponse:
            if p == 0:
                for i in xrange(aa._cache['jys'].shape[0]):
                    aa._cache['jys'][i,:] *= Bij_sf(aa._cache['s_top'][i,:], pol)
            else:
                for i in xrange(aa._cache['jys'].shape[0]):
                    aa._cache['jys'][i,:] /= Bij_sf(aa._cache['s_top'][i,:], pols[p-1])	# Remove the old pol
                    aa._cache['jys'][i,:] *=  Bij_sf(aa._cache['s_top'][i,:], pol)
                    
        ## Simulate
        if not flatResponse:
            uvw1, vis1 = FastVis(aa, baselines, chanMin, chanMax, pcAz, pcEl, resolve_src=resolve_src)
        else:
            currentVars = locals().keys()
            if 'uvw1' not in currentVars or 'vis1' not in currentVars:
                uvw1, vis1 = FastVis(aa, baselines, chanMin, chanMax, pcAz, pcEl, resolve_src=resolve_src)
                
        ## Unpack the data and repack it into a data dictionary
        UVData['bls'][pol] = []
        UVData['uvw'][pol] = []
        UVData['vis'][pol] = []
        UVData['wgt'][pol] = []
        UVData['msk'][pol] = []
        UVData['jd'][pol] = []
        
        if mask is None:
            for i in xrange(uvw1.shape[0]):
                UVData['bls'][pol].append( baselines[i] )
                UVData['uvw'][pol].append( uvw1[i,:,:] )
                UVData['vis'][pol].append( vis1[i,:] )
                UVData['wgt'][pol].append( numpy.ones_like(vis1[i,:])*len(UVData['vis'][pol][-1]) )
                UVData['msk'][pol].append( numpy.zeros_like(UVData['vis'][pol][-1]) )
                UVData['jd'][pol].append( jd )
                
        else:
            for i in xrange(uvw1.shape[0]):
                msk = mask[i]
                
                UVData['bls'][pol].append( baselines[i] )
                UVData['uvw'][pol].append( uvw1[i,:,:].compress(numpy.logical_not(msk), axis=2) )
                UVData['vis'][pol].append( vis1[i,:].compress(numpy.logical_not(msk)) )
                UVData['wgt'][pol].append( numpy.ones_like(UVData['vis'][pol][-1])*len(UVData['vis'][pol][-1]) )
                UVData['msk'][pol].append( msk )
                UVData['jd'][pol].append( jd )
                
    # Cleanup
    if not flatResponse:
        for i in xrange(aa._cache['jys'].shape[0]):
            aa._cache['jys'][i,:] /= Bij_sf(aa._cache['s_top'][i,:], pols[-1])	# Remove the old pol
            aa._cache['jys'][i,:] /= Gij_sf								# Remove the bandpass
            
    # Return
    return UVData


def buildSimData(aa, srcs, pols=['xx', 'yy', 'xy', 'yx'], jd=None, chan=None, phaseCenter='z', baselines=None, mask=None,  flatResponse=False, resolve_src=False, verbose=False):
    """
    Given an AIPY AntennaArray object and a dictionary of sources from 
    aipy.src.get_catalog, returned a data dictionary of simulated data taken at 
    zenith.  Optinally, the data can be masked using some referenced (observed) 
    data set or only a specific sub-set of baselines.
    
    .. versionchanged:: 1.0.1
        * Added a 'flatResponse' keyword to make it easy to toggle on and off
          the spectral and spatial response of the array for the simulation
        * Added a 'resolve_src' keyword to turn on source resolution effects
        
    ..versionchanged:: 0.4.0
        Added the 'pols' keyword to only compute certain polarization components
    """
    
    # Update the JD if necessary
    if jd is not None:
        try:
            junk = len(jd)
        except TypeError:
            jd = [jd]
    else:
        jd = [aa.get_jultime()]
        
    # Build up output dictionary
    freq = aa.get_afreqs()*1e9
    if len(freq.shape) == 2:
        freq = freq[0,:]
    UVData = {'freq': freq, 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
    if mask is None:
        UVData['isMasked'] = False
    else:
        UVData['isMasked'] = True
    
    # Loop over polarizations to populate the simulated data set
    for pol in pols:
        UVData['bls'][pol] = []
        UVData['uvw'][pol] = []
        UVData['vis'][pol] = []
        UVData['wgt'][pol] = []
        UVData['msk'][pol] = []
        UVData['jd'][pol] = []

    # Loop over Julian days to fill in the simulated data set
    jdCounter = 1
    for juldate in jd:
        oBlk = __buildSimData(aa, srcs, pols=pols, jd=juldate, chan=chan, phaseCenter=phaseCenter, baselines=baselines, mask=mask, verbose=verbose, count=jdCounter, max=len(jd), flatResponse=flatResponse, resolve_src=resolve_src)
        jdCounter = jdCounter + 1

        for pol in oBlk['bls']:
            for b,u,v,w,m,j in zip(oBlk['bls'][pol], oBlk['uvw'][pol], oBlk['vis'][pol], oBlk['wgt'][pol], oBlk['msk'][pol], oBlk['jd'][pol]):
                UVData['bls'][pol].append( b )
                UVData['uvw'][pol].append( u )
                UVData['vis'][pol].append( v )
                UVData['wgt'][pol].append( w )
                UVData['msk'][pol].append( m )
                UVData['jd'][pol].append( j )
                
    return UVData


def scaleData(dataDict, amps, delays, phaseOffsets=None):
    """
    Apply a set of antenna-based real gain values and phase delays in ns to a 
    data dictionary.  Returned the new scaled and delayed dictionary.
    
    ..versionchanged:: 0.6.3
        Added a keyword so that phase offsets (in radians) can also be specified
    ..versionchanged:: 0.4.0
        The delays are now expected to be in nanoseconds rather than radians.
    """
    
    # Build the data dictionary to hold the scaled and delayed data
    sclUVData = {'freq': (dataDict['freq']).copy(), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
    if dataDict['isMasked']:
        sclUVData['isMasked'] = True
    else:
        sclUVData['isMasked'] = False
    fq = dataDict['freq'] / 1e9
    
    if phaseOffsets is None:
        phaseOffsets = numpy.zeros_like(delays)
        
    cGains = []
    for i in xrange(len(amps)):
        cGains.append( amps[i]*numpy.exp(2j*numpy.pi*fq*delays[i] + 1j*phaseOffsets[i]) )
        
    # Apply the scales and delays for all polarization pairs found in the original data
    for pol in dataDict['vis'].keys():
        sclUVData['bls'][pol] = []
        sclUVData['uvw'][pol] = []
        sclUVData['vis'][pol] = []
        sclUVData['wgt'][pol] = copy.copy(dataDict['wgt'][pol])
        sclUVData['msk'][pol] = copy.copy(dataDict['msk'][pol])
        sclUVData['jd'][pol] = copy.copy(dataDict['jd'][pol])

        for (i,j),uvw,vis in zip(dataDict['bls'][pol], dataDict['uvw'][pol], dataDict['vis'][pol]):
            sclUVData['bls'][pol].append( (i,j) )
            sclUVData['uvw'][pol].append( uvw )
            sclUVData['vis'][pol].append( vis*cGains[j].conj()*cGains[i] )
            
    return sclUVData
    

def shiftData(dataDict, aa):
    """
    Shift the uvw coordinates in one data dictionary to a new set of uvw 
    coordinates that correspond to a new AntennaArray object.  This is useful
    for looking at how positional errors in the array affect the data.
    """
    
    # Build the data dictionary to hold the shifted data
    shftData = {'freq': dataDict['freq'].copy(), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
    if dataDict['isMasked']:
        shftData['isMasked'] = True
    else:
        shftData['isMasked'] = False
        
    # Apply the coordinate shift all polarization pairs found in the original data
    for pol in dataDict['vis'].keys():
        shftData['bls'][pol] = copy.copy(dataDict['bls'][pol])
        shftData['uvw'][pol] = []
        shftData['vis'][pol] = copy.copy(dataDict['vis'][pol])
        shftData['wgt'][pol] = copy.copy(dataDict['wgt'][pol])
        shftData['msk'][pol] = copy.copy(dataDict['msk'][pol])
        shftData['jd'][pol] = copy.copy(dataDict['jd'][pol])

        for (i,j),m in zip(dataDict['bls'][pol], dataDict['msk'][pol]):
            crds = aa.gen_uvw(j, i, src='z')[:,0,:]
            if dataDict['isMasked']:
                crds = crds.compress(numpy.logical_not(numpy.logical_not(m), axis=2))
            shftData['uvw'][pol].append( crds )
            
    return shftData


def addBaselineNoise(dataDict, SEFD, tInt, bandwidth=None, efficiency=1.0):
    """
    Given a data dictionary of visibilities, an SEFD or array SEFDs in Jy, 
    and an integration time in seconds, add noise to the visibilities 
    assuming that the "weak source" limit.
    
    This function implements Equation 9-15 from Chapter 9 of "Synthesis 
    Imaging in Radio Astronomy II".
    
    .. versionadded:: 1.0.2
    """
    
    # Figure out the bandwidth from the frequency list in the 
    if bandwidth is None:
        try:
            bandwidth = dataDict['freq'][1] - dataDict['freq'][0]
        except IndexError:
            raise RuntimeError("Too few frequencies to determine the bandwidth, use the 'bandwidth' keyword")
            
    # Calculate the standard deviation of the real/imaginary noise
    visNoiseSigma = SEFD / efficiency / numpy.sqrt(2.0*bandwidth*tInt)
    
    # Make sure that we have an SEFD for each antenna
    ants = []
    for pol in dataDict['bls'].keys():
        for i,j in dataDict['bls'][pol]:
            if i not in ants:
                ants.append( i )
            if j not in ants:
                ants.append( j )
    nAnts = len(ants)
    try:
        if len(SEFD) != nAnts:
            raise RuntimeError("Mis-match between the number of SEFDs supplied and the number of antennas in the data dictionary")
            
    except TypeError:
        SEFD = numpy.ones(nAnts)*SEFD
        
    # Build the data dictionary to hold the data with noise added
    bnData = {'freq': dataDict['freq'].copy(), 'uvw': {}, 'vis': {}, 'wgt': {}, 'msk': {}, 'bls': {}, 'jd': {}}
    if dataDict['isMasked']:
        bnData['isMasked'] = True
    else:
        bnData['isMasked'] = False
        
    # Copy the original data over and add in the noise
    for pol in dataDict['vis'].keys():
        bnData['bls'][pol] = copy.copy(dataDict['bls'][pol])
        bnData['uvw'][pol] = copy.copy(dataDict['uvw'][pol])
        bnData['vis'][pol] = []
        bnData['wgt'][pol] = copy.copy(dataDict['wgt'][pol])
        bnData['msk'][pol] = copy.copy(dataDict['msk'][pol])
        bnData['jd'][pol] = copy.copy(dataDict['jd'][pol])
        
        ## Apply this to the visibilities
        for bl in xrange(len(dataDict['vis'][pol])):
            vShape = dataDict['vis'][pol][bl].shape
            
            ### Calculate the expected noise
            visNoise = visNoiseSigma * (numpy.random.randn(*vShape) + 1j*numpy.random.randn(*vShape))
            
            ### Apply
            bnData['vis'][pol].append( dataDict['vis'][pol][bl] + visNoise )
            
    return bnData



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    long = int
    
"""
Module to write VDIF frames.  The implementation of this module is similar
to that of lsl.sim.tbw in that the primary element defined in this module is
a Frame object which as attribute functions that can create a numpy 
representation of the raw frame and write that raw frame to an open file-
handle.

.. seealso::
    :mod:`lsl.sim.tbw`
"""

import numpy
import ephem

from lsl.common import dp as dp_common
import lsl.astro as astro

__version__ = '0.1'
__revision__ = '$Rev: 2577 $'
__all__ = ['Frame', '__version__', '__revision__', '__all__']

vdifEpoch = ephem.Date('2000/01/01 00:00:00.00')
unixEpoch = ephem.Date('1970/01/01 00:00:00.00')


class Frame(object):
    """
    Object to create and write a VDIF (VLBI Data Interchange Format) 
    frame (version 1, June 26, 2009).
    """
    
    def __init__(self, stand=0, time=0, bits=8, data=None, sampleRate=dp_common.fS):
        self.stand = stand
        self.time = time
        self.bits = bits
        self.data = numpy.squeeze(data)
        self.sampleRate = sampleRate
        if data.dtype.kind == 'c':
            self.dataReal = False
        else:
            self.dataReal = True
            
        self.seconds = 0
        self.frame = 0
        self.epoch = 0
        
        # Validate
        if self.bits not in (1, 2, 4, 8):
            raise ValueError("Unsupported output bitdepth: %i" % self.bits)
                    
        # Convert the time from UNIX to epoch and make the data ready to 
        # be written to the disk.
        self.__dataAdjust()
        self.__setEpoch()
        
    def __dataAdjust(self):
        """
        Adjust the data to the number of bits required for the frame 
        and convert to unsigned 32-bit integers.  Also, interlace complex
        arrays.
        """
        
        if self.dataReal:
            interlaced = self.data
        else:
            interlaced = numpy.zeros(2*len(self.data))
            interlaced[0::2] = self.data.real
            interlaced[1::2] = self.data.imag
        
        biased = interlaced + 2**(self.bits-1)
        biased = biased.astype(numpy.uint32) & (2**self.bits-1)
        
        self.data = biased
        
    def __setEpoch(self):
        """
        Convert a UNIX timestap in seconds since 1970 to seconds since
        January 1st, 2000 using ephem.Data objects.  Also, calculate the 
        frame number since the beginning of the second.
        """
        
        # UNIX time to seconds since DJD  = 0
        curEpoch = float(unixEpoch)*astro.SECS_IN_DAY + self.time
        self.epoch = 0
        # Seconds since the VDIF epoch
        epochSeconds = curEpoch - float(vdifEpoch)*astro.SECS_IN_DAY
        # Integer seconds
        self.seconds = long(epochSeconds)
        
        # Compute the frames since the beginning of the second
        frame = (epochSeconds - self.seconds) * (self.sampleRate/len(self.data))
        self.frame = long(round(frame))

    def createRawFrame(self):
        """
        Using the data and information stored in the object, create a
        numpy array on unsigned 1-byte integers that represents the frame.
        """
        
        # Find out how many samples can be packed into a work and build
        # the corresponding numpy array.
        samplesPerWord = int(32 // self.bits)
        raw = numpy.zeros(32 + 4*len(self.data)//samplesPerWord, dtype=numpy.uint8)

        # Valid data, standard (not legacy) 32-bit header, and seconds since 
        # the 01/01/2000 epoch.
        raw[3] = (0 << 7) | (0 << 6) | ((self.seconds >> 24) & 63)
        raw[2] = (self.seconds >> 16) & 255
        raw[1] = (self.seconds >> 8) & 255
        raw[0] = self.seconds & 255

        # Refernce epoch (0 == 01/01/2000) and frame count
        raw[7] = self.epoch & 63
        raw[6] = (self.frame >> 16) & 255
        raw[5] = (self.frame >> 8) & 255
        raw[4] = self.frame & 255

        # VDIF version number, number of channels (just 1), and data frame 
        # length in units to 8-bytes (8 raw array elements)
        raw[11] = (1 << 6) | (0 & 31)
        raw[10] = ((len(raw) // 8) >> 16) & 255
        raw[9] = ((len(raw) // 8) >> 8) & 255
        raw[8] = (len(raw) // 8) & 255

        # Data type, bits per sample, thread ID, and station ID
        # NB:  The thread ID is fixed at `0'
        if self.dataReal:
            raw[15] = (0 << 7) | ((self.bits-1 & 31) << 2) | ((0 >> 8) & 3)
        else:
            raw[15] = (1 << 7) | ((self.bits-1 & 31) << 2) | ((0 >> 8) & 3)
        raw[14] = 0 & 255
        raw[13] = (self.stand >> 8) & 255
        raw[12] = self.stand & 255
        
        # User-defined words 4 through 7 (array entries 16 to 31)
        ## NICT format so that we can put in the sample rate
        sampleRatekHz = int(self.sampleRate / 1000.0)
        raw[19] = 0x01
        raw[18] = (0 << 7) | ((sampleRatekHz >> 16) & 0xFF)
        raw[17] = (sampleRatekHz >> 8) & 0xFF
        raw[16] = sampleRatekHz & 0xFF
        raw[20:24] = 0
        raw[24:28] = 0
        raw[28:32] = 0
        
        # Data values
        for f in range(32, len(raw), 4):
            i = (f - 32) // 4
            word = 0
            for p in range(samplesPerWord):
                word |= (self.data[i*samplesPerWord + p] << (self.bits*p))
            raw[f+3] = (word >> 24) & 255
            raw[f+2] = (word >> 16) & 255
            raw[f+1] = (word >> 8) & 255
            raw[f+0] = word & 255
            
        return raw

    def writeRawFrame(self, fh):
        """
        Create a numpy representation of the VDIF frame and then write 
        it to the specified file handle.
        """
        
        rawFrame = self.createRawFrame()
        rawFrame.tofile(fh)
